"""Utility functions for handling and presenting numeric data."""
import itertools
import os
from typing import Union, List, Dict, Tuple, Optional, Set

import math
import matplotlib.pyplot as plt
import numpy as np
import sympy
from pandas import DataFrame, Series
from pddl_plus_parser.lisp_parsers import PDDLTokenizer
from pddl_plus_parser.models import Precondition, PDDLFunction, construct_expression_tree, NumericalExpressionTree
from scipy.spatial import ConvexHull, convex_hull_plot_2d
from sklearn.feature_selection import VarianceThreshold
from sklearn.linear_model import LinearRegression

from sam_learning.core.learning_types import ConditionType
from utilities.util_types import NUMERIC_PRECISION

DECIMAL_DIGITS = int(os.environ.get(NUMERIC_PRECISION, 4))
EPSILON = 10 ** (-DECIMAL_DIGITS)
ELIMINATION_THRESHOLD = 0.01


def get_num_independent_equations(data_matrix: DataFrame) -> int:
    """Calculates the number of independent equations in the given matrix.

    :param data_matrix: the matrix of the previous state values.
    :return: the number of independent equations.
    """
    values_matrix = data_matrix.to_numpy()
    num_rows = values_matrix.shape[0]
    values_matrix_with_bias = np.c_[values_matrix, np.ones(num_rows)]
    _, pivot_cols = sympy.Matrix(values_matrix_with_bias).rref()
    return len(pivot_cols)


def prettify_floating_point_number(number: float) -> float:
    """Converts the floating point number into a prettier form so that the created equations would be more presentable.

    :param number: the RAW number received from the learning process.
    :return: the prettified version of the number.
    """
    return int(number) if abs(number - int(number)) < EPSILON else number


def construct_multiplication_strings(coefficients_vector: Union[np.ndarray, List[float]], function_variables: List[str]) -> List[str]:
    """Constructs the strings representing the multiplications of the function variables with the coefficient.

    :param coefficients_vector: the coefficient that multiplies the function vector.
    :param function_variables: the name of the numeric fluents that are being used.
    :return: the representation of the fluents multiplied by the coefficients.
    """
    product_components = []
    for func, coefficient in zip(function_variables, coefficients_vector):
        if abs(round(coefficient, DECIMAL_DIGITS)) <= EPSILON:
            continue

        if func == "(dummy)":
            product_components.append(f"{prettify_floating_point_number(round(coefficient, DECIMAL_DIGITS))}")

        elif coefficient == 1.0:
            product_components.append(func)

        else:
            product_components.append(f"(* {func} {prettify_floating_point_number(round(coefficient, DECIMAL_DIGITS))})")

    return product_components


def prettify_coefficients(coefficients: List[float]) -> List[float]:
    """Converts the coefficients into a prettier form so that the created equations would be more presentable.

    :param coefficients: the RAW coefficients received from the linear regression.
    :return: the prettified version of the coefficients.
    """
    coefficients = [coef if abs(coef) > EPSILON else 0.0 for coef in coefficients]
    prettified_coefficients = [round(value, DECIMAL_DIGITS) for value in coefficients]
    return prettified_coefficients


def construct_projected_variable_strings(
    function_variables: List[str], shift_point: Union[np.ndarray, List[float]], projection_basis: Union[np.ndarray, List[List[float]]],
) -> List[str]:
    """Constructs the strings representing the multiplications of the function variables with the coefficient.

    :param function_variables: the name of the numeric fluents that are being used.
    :param shift_point: the point in which the data was shifted according to.
    :param projection_basis: the basis in which the data was projected to.
    :return: the new variable names after applying the PCA model transformation
    """
    shifted_by_mean = []
    for func, shift_value in zip(function_variables, shift_point):
        component_function = func if shift_value == 0.0 else f"(- {func} {prettify_floating_point_number(round(shift_value, DECIMAL_DIGITS))})"
        shifted_by_mean.append(component_function)

    sum_of_product_by_components = []
    for row in range(len(projection_basis)):
        product_by_components_row = []
        for shifted, component in zip(shifted_by_mean, projection_basis[row]):
            if abs(round(component, DECIMAL_DIGITS)) <= EPSILON:
                continue

            if component == 1.0:
                product_by_components_row.append(shifted)
                continue

            product_by_components_row.append(f"(* {shifted} {prettify_floating_point_number(round(component, DECIMAL_DIGITS))})")

        sum_of_product_by_components.append(construct_linear_equation_string(product_by_components_row))

    return sum_of_product_by_components


def construct_linear_equation_string(multiplication_parts: List[str]) -> str:
    """Construct the addition parts of the linear equation string.

    :param multiplication_parts: the multiplication function strings that are multiplied by the coefficient.
    :return: the string representing the sum of the linear variables.
    """
    if len(multiplication_parts) == 1:
        return multiplication_parts[0]

    inner_layer = construct_linear_equation_string(multiplication_parts[1:])
    return f"(+ {multiplication_parts[0]} {inner_layer})"


def construct_non_circular_assignment(lifted_function: str, coefficients_map: Dict[str, float], previous_value: float, next_value: float) -> str:
    """Changes circular assignment statements to be non-circular.

    Note:
        Since numeric solvers don't approve circular dependencies we need format the assignment operations to be
        in the form of increase / decrease.

    :param lifted_function: the assigned variable.
    :param coefficients_map: the calculated coefficient map.
    :param previous_value: the numeric value of the function prior to the action's execution.
    :param next_value: the numeric value of the function after the action's execution.
    :return: the formatted string without circular dependencies.
    """
    if previous_value < next_value:
        coefficients_map[lifted_function] = coefficients_map[lifted_function] - 1

    else:
        coefficients_map = {k: -v for k, v in coefficients_map.items()}
        coefficients_map[lifted_function] = coefficients_map[lifted_function] + 1

    multiplication_functions = construct_multiplication_strings(list(coefficients_map.values()), list(coefficients_map.keys()))
    constructed_right_side = construct_linear_equation_string(multiplication_functions)

    if previous_value < next_value:
        return f"(increase {lifted_function} {constructed_right_side})"

    return f"(decrease {lifted_function} {constructed_right_side})"


def extract_numeric_linear_coefficient(function1_values: Series, function2_values: Series) -> float:
    """Extract the first real numeric value from the two lists of values since there might be NaN values.

    :param function1_values: the values of the first function.
    :param function2_values: the values of the second function.
    :return: the first numeric divisor.
    """
    linear_coeff = 0
    denominator = np.where(function2_values == 0, 1e-9, function2_values)
    division_res = np.array(function1_values) / np.array(denominator)
    for value in division_res:
        if not math.isnan(value) and not math.isinf(value) and not math.isclose(value, 0.0, abs_tol=ELIMINATION_THRESHOLD):
            linear_coeff = value
            break

    return prettify_floating_point_number(round(linear_coeff, DECIMAL_DIGITS))


def filter_constant_features(input_df: DataFrame, columns_to_ignore: Optional[List[str]] = []) -> Tuple[DataFrame, List[str], List[str]]:
    """Filters out fluents that contain only constant values since they do not contribute to the convex hull.

    :param input_df: the matrix of the previous state values.
    :param columns_to_ignore: the list of columns that should be ignored.
    :return: the filtered matrix and the equality strings, i.e. the strings of the values that should be equal.
    """
    if len(input_df) == 1:
        return input_df, [], []

    equal_fluent_strs, removed_fluents = [], []
    relevant_columns = [col for col in input_df.columns if col not in columns_to_ignore]
    try:
        selector = VarianceThreshold().set_output(transform="pandas")
        result_df = selector.fit_transform(input_df[relevant_columns])
        filtered_columns = [column for column in relevant_columns if column not in result_df.columns]
        for col in filtered_columns:
            equal_fluent_strs.append(f"(= {col} {input_df[col].unique()[0]})")
            removed_fluents.append(col)

        return result_df, equal_fluent_strs, removed_fluents

    except ValueError:
        # no feature in X meets the variance threshold 0.0 (i.e. all features are constant)
        for col in input_df.columns:
            equal_fluent_strs.append(f"(= {col} {input_df[col].unique()[0]})")

        return DataFrame(), equal_fluent_strs, input_df.columns


def detect_linear_dependent_features(data_matrix: DataFrame, columns_to_ignore: List[str] = []) -> Tuple[DataFrame, List[str], Dict[str, str]]:
    """Detects linear dependent features and adds the equality constraints to the problem.

    The idea is: put together these column vectors as a matrix and calculate its row-echelon form.
    If the row-echelon form is diagonal with only ones, the set of vectors is independent, otherwise, it is dependent.

    :param data_matrix: the matrix of the previous state values.
    :param columns_to_ignore: the list of columns that are already composed of linear combinations of other columns and should be ignored.
    :return: the filtered matrix and the equality strings, i.e. the strings of the values that should be equal and the
        column to column mapping.
    """
    additional_conditions = []
    dependent_columns = {}

    if len(data_matrix) < 2:
        # cannot detect linear dependency with less than two samples
        return data_matrix, additional_conditions, dependent_columns

    data_matrix_copy = data_matrix[[col for col in data_matrix.columns.tolist() if col not in columns_to_ignore]].copy()
    # Ignoring constant columns since they add noise to the matrix
    data_matrix_copy = data_matrix_copy.loc[:, (data_matrix_copy != data_matrix_copy.iloc[0]).any()]

    for col1, col2 in itertools.combinations(data_matrix_copy.columns, 2):
        reduced_form, _ = sympy.Matrix(data_matrix_copy[[col1, col2]].values).rref()
        diagonal_required_result = np.array([[1, 0], [0, 1]])
        if np.array_equal(diagonal_required_result, np.diag(np.diag(reduced_form))):
            continue  # The columns are independent

        independent_column, dependent_column = col1, col2
        linear_coeff = extract_numeric_linear_coefficient(data_matrix_copy[dependent_column], data_matrix_copy[independent_column])
        if linear_coeff == 0:
            continue

        additional_conditions.append(f"(= {dependent_column} (* {linear_coeff} {independent_column}))")
        dependent_columns[dependent_column] = independent_column

    filtered_matrix = data_matrix[[col for col in data_matrix.columns.tolist() if col not in dependent_columns]]
    return filtered_matrix, additional_conditions, dependent_columns


def construct_numeric_conditions(
    conditions: List[str], condition_type: ConditionType, domain_functions: Dict[str, PDDLFunction]
) -> Optional[Precondition]:
    """Construct the numeric conditions from the given equality conditions.

    :param conditions: the condition strings to create the preconditions from.
    :param condition_type: the type of the conditions ('and' or 'or').
    :param domain_functions: the domain functions to use for the numeric conditions.
    :return: the constructed numeric precondition.
    """
    if len(conditions) == 0:
        return None

    precondition_type = "and" if condition_type == ConditionType.conjunctive else "or"
    constructed_precondition = Precondition(precondition_type)
    precondition_parts = []
    for condition in conditions:
        numeric_tokenizer = PDDLTokenizer(pddl_str=condition)
        tokens = numeric_tokenizer.parse()
        numeric_expression = construct_expression_tree(tokens, domain_functions)
        precondition_parts.append(NumericalExpressionTree(numeric_expression))

    for part in precondition_parts:
        constructed_precondition.add_condition(part)

    return constructed_precondition


def construct_numeric_effects(effects: List[str], domain_functions: Dict[str, PDDLFunction]) -> Set[NumericalExpressionTree]:
    """Construct the numeric effects for the given set of input strings.

    :param effects: the effect strings to create the numeric effects from.
    :param domain_functions: the domain functions to use for the numeric effects.
    :return: the constructed numeric precondition.
    """
    if len(effects) == 0:
        return set()

    numeric_effects = []
    for condition in effects:
        numeric_tokenizer = PDDLTokenizer(pddl_str=condition)
        tokens = numeric_tokenizer.parse()
        numeric_expression = construct_expression_tree(tokens, domain_functions)
        numeric_effects.append(numeric_expression)

    return {NumericalExpressionTree(expr) for expr in numeric_effects}


def extended_gram_schmidt(input_basis_vectors: List[List[float]], eigen_vectors: Optional[List[List[float]]] = None) -> List[List[float]]:
    """Runs the extended Gram-Schmidt algorithm on the input basis vectors.

    Note:
        The algorithm is extended in the that it can handle the additional input eigen vectors and return the
        orthonormal basis vectors of the input basis vectors and eigen vectors.

    :param input_basis_vectors: The input basis vectors.
    :param eigen_vectors: The eigen vectors - optional and can be none.
    :return: The orthonormal basis vectors of the input basis vectors and eigen vectors.
    """
    non_normal_vectors = eigen_vectors.copy() if eigen_vectors else []
    normal_vectors = []
    for vector in input_basis_vectors:
        # Gram Schmidt magic
        projected_vector = vector - np.sum([(np.dot(vector, b) / np.linalg.norm(b) ** 2) * np.array(b) for b in non_normal_vectors], axis=0)
        if not (np.absolute(projected_vector) > EPSILON).any():
            continue

        non_normal_vectors.append(projected_vector.tolist())
        normal_vectors.append((projected_vector / np.linalg.norm(projected_vector)).tolist())

    return normal_vectors


def display_convex_hull_2d(action_name: str, hull: ConvexHull) -> None:
    """Displays the convex hull in 2D as a plot.

    :param action_name: the name of the action with its convex hull displayed.
    :param hull: the convex hull to display.
    """
    fig = convex_hull_plot_2d(hull)
    ax = fig.gca()
    ax.set_title(f"{action_name} - convex hull")
    plt.show()


def display_convex_hull_3d(action_name: str, hull: ConvexHull) -> None:
    """Displays the convex hull in 3D as a plot.

    :param action_name: the name of the action with its convex hull displayed.
    :param hull: the convex hull to display.
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set_title(f"{action_name} - convex hull")
    for simplex in hull.simplices:
        simplex = np.append(simplex, simplex[0])  # Repeat the first point to create a closed shape
        ax.plot(hull.points[simplex, 0], hull.points[simplex, 1], hull.points[simplex, 2], "b-")

    plt.show()


def display_convex_hull(action_name: str, display_mode: bool, hull: ConvexHull) -> None:
    """Displays the convex hull in as a plot.

    :param action_name: the name of the action with its convex hull displayed.
    :param display_mode: whether to display the plot.
    :param hull: the convex hull to display.
    """
    if not display_mode:
        return

    dimensionality = hull.points.shape[1]
    if dimensionality == 2:
        display_convex_hull_2d(action_name, hull)

    elif dimensionality == 3:
        display_convex_hull_3d(action_name, hull)


def create_monomials(domain_functions: List[str], polynom_degree: int = 0) -> List[List[str]]:
    """Creates the monomials from the state fluents.

    :return: the monomials from the state fluents.
    """
    monomials = list([item] for item in domain_functions)
    if polynom_degree == 0:
        return monomials

    if polynom_degree == 1:
        for first_fluent, second_fluent in itertools.combinations(domain_functions, r=2):
            monomial = sorted([first_fluent, second_fluent])
            monomials.append(monomial)

    else:
        for degree in range(2, polynom_degree + 1):
            for fluent_combination in itertools.combinations_with_replacement(domain_functions, r=degree):
                monomial = sorted(list(fluent_combination))
                monomials.append(monomial)

    return monomials


def create_grounded_monomials(monomials: List[List[str]], state_fluents: Dict[str, PDDLFunction]) -> Dict[str, float]:
    """Creates a dictionary of grounded monomials based on given monomials and state fluents.

    This function iterates over a list of monomials, where each monomial is a list of
    strings representing components. It computes the product of the values
    of the given `state_fluents` for each monomial and stores the result
    in a dictionary with the polynomial string representation of the monomial as its key.

    :param monomials: A list of monomials, where each monomial is a list of strings
                      representing components.
    :param state_fluents: A dictionary where keys are strings representing fluents,
                          and values are instances of PDDLFunction, containing their
                          respective values.
    :return: A dictionary where keys are the string representations of monomials
             and values are the computed product of the respective fluent values.
    """
    sample_dataset = {}
    for monomial in monomials:
        if any([component not in state_fluents for component in monomial]):
            continue

        sample_dataset[create_polynomial_string(monomial)] = np.prod([state_fluents[fluent].value for fluent in monomial])

    return sample_dataset


def _create_polynomial_string_recursive(fluents: List[str]) -> str:
    """Creates the polynomial string representing the equation recursively.

    :param fluents: the numeric fluents to create the polynomial string from.
    :return: the polynomial string representing the equation.
    """
    if len(fluents) == 1:
        return fluents[0]

    return f"(* {fluents[0]} {_create_polynomial_string_recursive(fluents[1:])})"


def create_polynomial_string(fluents: List[str]) -> str:
    """The auxiliary function that creates the polynomial string representing the equation.

    :param fluents: the numeric fluents to create the polynomial string from.
    :return: the polynomial string representing the equation.
    """
    return _create_polynomial_string_recursive(fluents)


def divide_span_by_common_denominator(equations_list: List[List[float]]) -> List[List[float]]:
    """Divides the span by the common denominator.

    :param equations_list: the list of equations (all equal to zero) that can be divided by the common denominator.
    :return: the span divided by the common denominator.
    """
    new_span = []
    for equation in equations_list:
        common_denominator = [coeff for coeff in equation if coeff != 0][0]
        new_span.append([coeff / common_denominator for coeff in equation])

    return new_span


def _first_non_zero_index(numbers_list: List[float]) -> int:
    """

    :param numbers_list:
    :return:
    """
    for index, value in enumerate(numbers_list):
        if value != 0:
            return index
    return -1


def reduce_complementary_conditions_from_convex_hull(convex_hull: List[List[float]], complementary_basis: List[List[float]]) -> List[List[float]]:
    """Reduces the complementary conditions from the Gram-Schmidt basis.

    :param convex_hull: the basis of the projection.
    :param complementary_basis: the complementary basis to the projection.
    :return: the reduced complementary conditions.
    """
    if len(complementary_basis) == 0:
        return convex_hull

    reduced_convex_hull = np.array(convex_hull).copy()
    for complementary_vector in complementary_basis:
        non_zero_index = _first_non_zero_index(complementary_vector)
        if non_zero_index == -1:
            continue

        new_conditions = [-1 / complementary_vector[non_zero_index] * complementary_vector[i] for i in range(len(complementary_vector))]
        reduced_convex_hull = (
            reduced_convex_hull
            + np.tile(np.array(new_conditions), (reduced_convex_hull.shape[0], 1)) * reduced_convex_hull[:, non_zero_index][:, np.newaxis]
        )
        reduced_convex_hull[:, non_zero_index] = 0

    return reduced_convex_hull.tolist()


def remove_complex_linear_dependencies(data: DataFrame) -> Tuple[DataFrame, List[str]]:
    """Removes the complex linear dependencies from the data matrix.

    :param data: the matrix of the previous state values.
    :return: the filtered matrix and the removed columns.
    """
    if len(data) == 0 or len(data.columns) == 1:
        return data, []

    constant_columns = data.columns[data.nunique() == 1].tolist()
    removed_columns = [*constant_columns]
    conditions = [f"(= {col} {data[col].unique()[0]})" for col in constant_columns]
    if len(data.columns) == len(constant_columns):
        return DataFrame(), conditions

    for feature_to_check in [col for col in data.columns if col not in removed_columns]:
        features = data[[col for col in data.columns if col != feature_to_check and col not in removed_columns]]
        if len(features.columns) == 0:
            # no columns left to check
            break

        model = LinearRegression()
        model.fit(features, data[feature_to_check])
        score = model.score(features, data[feature_to_check])
        if score != 1:
            # could not fit the feature to the other features - not linearly dependent
            continue

        coefficients = list(model.coef_) + [model.intercept_]
        coefficients = prettify_coefficients(coefficients)

        removed_columns.append(feature_to_check)
        if all([coef == 0 for coef in coefficients]):
            # the feature is a constant equal to zero
            conditions.append(f"(= {feature_to_check} 0)")
            continue

        multiplication_functions = construct_linear_equation_string(
            construct_multiplication_strings(coefficients, [*features.columns.tolist(), "(dummy)"])
        )
        conditions.append(f"(= {feature_to_check} {multiplication_functions})")

    return data.drop(columns=removed_columns), conditions


def construct_pddl_inequality_scheme(
    coefficient_matrix: np.ndarray, border_points: Union[np.ndarray, float], headers: List[str], sign_to_use: str = "<="
) -> List[str]:
    """Construct the inequality strings in the appropriate PDDL format.

    :param coefficient_matrix: the matrix containing the coefficient vectors for each inequality.
    :param border_points: the convex hull point which ensures that Ax <= b.
    :param headers: the headers of the columns in the coefficient matrix.
    :param sign_to_use: the sign to use in the inequalities (for cases when we want to use equalities).
    :return: the inequalities PDDL formatted strings.
    """
    inequalities = []
    if isinstance(border_points, float):
        # there is only one data point
        multiplication_functions = construct_multiplication_strings(coefficient_matrix, headers)
        constructed_left_side = construct_linear_equation_string(multiplication_functions)
        return [f"({sign_to_use} {constructed_left_side} {border_points})"]

    for inequality_coefficients, border_point in zip(coefficient_matrix, border_points):
        multiplication_functions = construct_multiplication_strings(inequality_coefficients, headers)
        constructed_left_side = construct_linear_equation_string(multiplication_functions)
        inequalities.append(f"({sign_to_use} {constructed_left_side} {border_point})")

    return inequalities



"""Module to validate the correctness of the learned action models that were generated."""
import csv
from pathlib import Path
from typing import Optional

from utilities import SolutionOutputTypes
from validators.safe_domain_validator import DomainValidator, DEBUG_STATISTICS

ACTION_LINE_REGEX = r"(\[\d+, \d+\]): \(.*\)"

ONLINE_SOLVING_STATISTICS = [
    "learning_algorithm",
    "episode_number",
    "number_of_steps_in_episode",
    "training_goal_achieved",
    "solver",
    "ok",
    "no_solution",
    "timeout",
    "solver_error",
    "not_applicable",
    "goal_not_achieved",
    "problems_ok",
    "problems_no_solution",
    "problems_timeout",
    "problems_solver_error",
    "problems_not_applicable",
    "problems_goal_not_achieved",
    "problems_validated_against_expert_plan",
    "problems_not_validated_against_expert_plan",
    "percent_ok",
    "percent_no_solution",
    "percent_timeout",
    "percent_solver_error",
    "percent_not_applicable",
    "percent_goal_not_achieved",
]


class OnlineLearningDomainValidator(DomainValidator):
    """Extension to th domain validator to support information relevant only to the online learning process."""

    def validate_domain(
            self, tested_domain_file_path: Path, test_set_directory_path: Optional[Path] = None,
            episode_number: Optional[int] = None, num_steps: Optional[int] = None,
            num_training_goal_achieved: int = 0, tolerance: float = 0.01) -> None:
        """Validates that using the input domain problems can be solved.

        :param tested_domain_file_path: the path of the domain that was learned using POL.
        :param test_set_directory_path: the path to the directory containing the test set problems.
        :param episode_number: the number of the current episode.
        :param num_steps: the number of steps that were taken in the current episode.
        :param num_training_goal_achieved: the number of times the training goal was achieved in the batch.
        :param tolerance: the tolerance to use when solving the problems.
        """
        self.logger.info("Solving the test set problems using the learned domain!")
        solving_report = self.solver.execute_solver(
            problems_directory_path=test_set_directory_path,
            domain_file_path=tested_domain_file_path,
            problems_prefix=self.problem_prefix,
            tolerance=tolerance
        )

        solving_stats = {solution_type.name: 0 for solution_type in SolutionOutputTypes}
        for debug_statistic in DEBUG_STATISTICS:
            solving_stats[debug_statistic] = []

        for problem_file_name, entry in solving_report.items():
            problem_file_path = test_set_directory_path / f"{problem_file_name}.pddl"
            if entry == SolutionOutputTypes.ok.name:
                solution_file_path = test_set_directory_path / f"{problem_file_name}.solution"
                self._validate_solution_content(solution_file_path=solution_file_path,
                                                problem_file_path=problem_file_path,
                                                iteration_statistics=solving_stats)

                continue

            solving_stats[entry] += 1
            solving_stats[f"problems_{entry}"].append(problem_file_name)

        self._calculate_solving_percentages(solving_stats)
        self.solving_stats.append({
            "learning_algorithm": self.learning_algorithm.name,
            "episode_number": episode_number,
            "number_of_steps_in_episode": num_steps,
            "training_goal_achieved": num_training_goal_achieved,
            "solver": self._solver_name,
            **solving_stats
        })
        self._clear_plans(test_set_directory_path)

    def write_statistics(self, fold_num: int) -> None:
        """Writes the statistics of the learned model into a CSV file.

        :param fold_num: the index of the fold that is currently being tested.
        """
        output_statistics_path = self.results_dir_path / f"online_nsam_{self.learning_algorithm.name}" \
                                                         f"_problem_solving_stats_{fold_num}.csv"
        with open(output_statistics_path, 'wt', newline='') as csv_file:
            test_set_writer = csv.DictWriter(csv_file, fieldnames=ONLINE_SOLVING_STATISTICS)
            test_set_writer.writeheader()
            test_set_writer.writerows(self.solving_stats)

    def write_complete_joint_statistics(self) -> None:
        """Writes a statistics file containing all the folds combined data."""
        output_path = self.results_dir_path / f"online_nsam_{self.learning_algorithm.name}_all_folds_solving_stats.csv"
        with open(output_path, 'wt', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=ONLINE_SOLVING_STATISTICS)
            writer.writeheader()
            writer.writerows(self.aggregated_solving_stats)



import json
import pathlib
import signal
import string
import subprocess
import sys
import time

from typing import Dict, Any

JOB_ID_MESSAGE = 'echo -e "\\nSLURM_JOBID:\\t\\t" $SLURM_JOBID'
JOB_NODELIST_MESSAGE = 'echo -e "SLURM_JOB_NODELIST:\\t" $SLURM_JOB_NODELIST "\\n\\n"'

FIRST_BREAKPOINT = 10
EXPERIMENTS_CONFIG_STR = "experiment_configurations"


def sigint_handler(sig_num, frame):
    signal.signal(signal.SIGINT, sigint_handler)
    print("\nCtrl-C pressed. Do you want to quit? (y/n): ", end="")
    response = input().strip().lower()
    if response in ["y", "yes"]:
        print("Deleting temp.sbatch file.")
        pathlib.Path("temp.sbatch").unlink()
        exit(0)


def progress_bar(progress, total):
    percent = 100 * (progress / float(total))
    percent_to_show = int(percent) // 2
    bar = "❚" * percent_to_show + "-" * (50 - percent_to_show)
    print(f"\r|{bar}| {percent:.2f}%", end="\r")


def write_sbatch_and_submit_job(sbatch_code: str):
    with open("temp.sbatch", "w+", newline="\n") as output_file:
        output_file.write(sbatch_code)

    data = subprocess.check_output(["sbatch", "temp.sbatch", "--parsable"]).decode()
    time.sleep(1)
    return int(data.split()[-1])


def get_configurations() -> Dict[str, Any]:
    print("Reading the configuration file.")
    configuration_file_path = sys.argv[1]
    with open(configuration_file_path, "rt") as configuration_file:
        return json.load(configuration_file)


def get_environment_variables(configurations: Dict[str, Any]) -> Dict[str, Any]:
    environment_variables_path = configurations["environment_variables_file_path"]
    with open(environment_variables_path, "rt") as environment_variables_file:
        return json.load(environment_variables_file)


def validate_job_running(sid: int):
    job_exists_command = ["squeue", "--job", f"{sid}"]
    try:
        result = subprocess.check_output(job_exists_command, shell=True).decode()
        if str(sid) in result:
            return sid

        return None

    except subprocess.CalledProcessError:
        return None


def setup_experiments_folds_job(
    code_directory, environment_variables, experiment, internal_iterations, experiment_size, should_create_random_trajectories: bool = True
):
    print(f"Working on the experiment with domain {experiment['domain_file_name']}\n")
    fold_creation_sid = submit_job(
        conda_env="online_nsam",
        mem="4G",
        python_file=f"{code_directory}/folder_creation_for_parallel_execution.py",
        jobname=f"create_folds_job_{experiment['domain_file_name']}",
        suppress_output=False,
        arguments=[
            f"--working_directory_path {experiment['working_directory_path']}",
            f"--domain_file_name {experiment['domain_file_name']}",
            f"--learning_algorithms {','.join([str(e) for e in experiment['compared_versions']])}",
            f"--internal_iterations {','.join([str(e) for e in internal_iterations])}",
            f"--problem_prefix {experiment['problems_prefix']}",
            f"--experiment_size {experiment_size}",
            f"--no_random_trajectories" if not should_create_random_trajectories else "",
        ],
        environment_variables=environment_variables,
        logs_directory=pathlib.Path(experiment["working_directory_path"]) / "logs",
    )
    print(f"Submitted job with sid {fold_creation_sid}\n")
    time.sleep(1)
    print("Removing the temp.sbatch file")
    pathlib.Path("temp.sbatch").unlink()
    return fold_creation_sid


def create_execution_arguments(experiment, fold, compared_version):
    arguments = []
    arguments.append(f"--fold_number {fold}")
    arguments.append(f"--learning_algorithm {compared_version}")
    for key, value in experiment.items():
        if key != "compared_versions" and key != "parallelization_data":
            arguments.append(f"--{key} {value}")

    return arguments


def create_internal_iterations_list(parallelization_data):
    max_train_size = (
        int(parallelization_data["experiment_size"] * 0.8) + 1 if "experiment_size" in parallelization_data else parallelization_data["max_index"] + 1
    )
    if parallelization_data["hop"] == 100:
        return [1] + list(range(10, 100, 10)) + list(range(100, max_train_size, 100))

    return list(range(1, FIRST_BREAKPOINT)) + list(range(FIRST_BREAKPOINT, max_train_size, parallelization_data["hop"]))


def create_experiment_folders(code_directory, environment_variables, experiment, should_create_random_trajectories: bool = True):
    print(f"Creating the directories containing the folds datasets for the experiments.")
    parallelization_data = experiment["parallelization_data"]
    internal_iterations = create_internal_iterations_list(parallelization_data)
    print(f"Internal iterations: {internal_iterations}")
    sid = setup_experiments_folds_job(
        code_directory=code_directory,
        environment_variables=environment_variables,
        experiment=experiment,
        internal_iterations=internal_iterations,
        experiment_size=parallelization_data.get("experiment_size", 100),
        should_create_random_trajectories=should_create_random_trajectories,
    )
    return internal_iterations, sid


def create_all_experiments_folders(code_directory, environment_variables, configurations, should_create_random_trajectories: bool = True):
    print("Creating the directories containing the folds datasets for the experiments.")
    jobs_sids = []
    output_internal_iterations = []
    for experiment_index, experiment in enumerate(configurations[EXPERIMENTS_CONFIG_STR]):
        internal_iterations, fold_creation_sid = create_experiment_folders(
            code_directory, environment_variables, experiment, should_create_random_trajectories
        )
        jobs_sids.append(fold_creation_sid)
        output_internal_iterations.append(internal_iterations)
        print(
            f"Submitted fold datasets folder creation job with the id {fold_creation_sid} for the experiment with domain {experiment['domain_file_name']}\n"
        )

    while any([validate_job_running(sid) is not None for sid in jobs_sids]):
        print("Waiting for the fold creation jobs to finish.")
        time.sleep(5)

    print("Finished building the experiment folds!")
    return output_internal_iterations


def submit_job(
    dependency=None,
    mem=None,
    runtime=None,
    jobname=None,
    cpus_per_task=None,
    suppress_error=False,
    error_file=None,
    suppress_output=False,
    output_file=None,
    conda_env=None,
    python_file=None,
    arguments=None,
    environment_variables=None,
    logs_directory=None,
):
    with open("experiments/cluster_scripts/template.sbatch", "rt", newline="\n") as template_file:
        text = template_file.read()
        sbatch_template = string.Template(text)

    # Complete the template code with the correct values
    all_environment_variables = [
        "export LOCAL_LOGS_PATH=/scratch/${SLURM_JOB_USER}/${SLURM_JOB_ID}",
        *[f"export {env_variable}={value}" for env_variable, value in environment_variables.items()],
    ]
    template_mapping = {
        "job_name": jobname if jobname else "job",
        "cpus_per_task": cpus_per_task if cpus_per_task else 1,
        "dependency_exists": "#" if not dependency else "",
        "dependency": dependency or "",
        "mem": mem or "8G",
        "conda_env": conda_env or "online_nsam",
        "script": python_file,
        "arguments": " ".join(arguments) if arguments else "",
        "environment_variables": "\n".join(all_environment_variables) if environment_variables else "",
        "job_info_print": JOB_ID_MESSAGE + "\n" + JOB_NODELIST_MESSAGE,
        "cluster_temp_logs_path": "$LOCAL_LOGS_PATH",
        "logs_dir": logs_directory or "/dev/null",
        "error": "#" if suppress_error else "",
    }

    sbatch_code = sbatch_template.substitute(template_mapping)
    try:
        return write_sbatch_and_submit_job(sbatch_code)

    except subprocess.CalledProcessError as e:
        template_mapping["dependency_exists"] = "#"  # Remove the dependency if it exists
        template_mapping["dependency"] = ""  # Remove the dependency if it exists
        sbatch_code = sbatch_template.substitute(template_mapping)
        return write_sbatch_and_submit_job(sbatch_code)


def submit_job_and_validate_execution(
    code_directory, configurations, experiment, fold, arguments, environment_variables, job_name, fold_creation_sid=None, python_file=None
):
    dependency_argument = None if not fold_creation_sid else f"afterok:{fold_creation_sid}"
    sid = submit_job(
        conda_env="online_nsam",
        mem="16G",
        python_file=python_file or f"{code_directory}/{configurations['experiments_script_path']}",
        jobname=job_name,
        dependency=dependency_argument,
        suppress_output=False,
        arguments=arguments,
        environment_variables=environment_variables,
        logs_directory=pathlib.Path(experiment["working_directory_path"]) / "logs",
        suppress_error=True,
    )
    time.sleep(2)
    return validate_job_running(sid)



import itertools
import logging
from collections import defaultdict
from itertools import permutations
from typing import List, Tuple, Dict, Set, Union, Optional

from pddl_plus_parser.models import Predicate, PDDLObject, GroundedPredicate, PDDLType, Domain, PDDLFunction, ActionCall, Action

from sam_learning.core.learner_domain import LearnerDomain, LearnerAction


def choose_objects_subset(array: List[str], subset_size: int) -> List[Tuple[str]]:
    """Choose r items our of a list size n.

    :param array: the input list.
    :param subset_size: the size of the subset.
    :return: a list containing subsets of the original list.
    """
    return list(permutations(array, subset_size))


class VocabularyCreator:
    """Creates predicate vocabulary from a domain containing action signatures and predicate definitions."""

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def _validate_type_matching(
        self, grounded_signatures: Dict[str, PDDLType], lifted_variable_to_match: Union[Predicate, PDDLFunction, Action, LearnerAction]
    ) -> bool:
        """Validates that the types of the grounded signature match the types of the predicate signature.

        :param grounded_signatures: the grounded predicate signature.
        :param lifted_variable_to_match: the lifted predicate.
        :return: whether the types match.
        """

        if len(grounded_signatures) != len(lifted_variable_to_match.signature):
            self.logger.debug(f"The number of objects - {grounded_signatures} does not match {lifted_variable_to_match.name}")
            return False

        for object_name, predicate_parameter in zip(grounded_signatures, lifted_variable_to_match.signature):
            parameter_type = lifted_variable_to_match.signature[predicate_parameter]
            grounded_type = grounded_signatures[object_name]
            if not grounded_type.is_sub_type(parameter_type):
                self.logger.debug(f"The combination of objects - {grounded_signatures}" f" does not fit {lifted_variable_to_match.name}'s signature")
                return False

        return True

    # TODO: vocabulary to allow duplicates in grounded_signature for future use
    def create_vocabulary(self, domain: Union[LearnerDomain, Domain], observed_objects: Dict[str, PDDLObject]) -> Dict[str, Set[GroundedPredicate]]:
        """Create a vocabulary of random combinations of the predicates parameters and objects.

        :param domain: the domain containing the predicates and the action signatures.
        :param observed_objects: the objects that were observed in the trajectory.
        :return: list containing all the predicates with the different combinations of parameters.
        """
        vocabulary = defaultdict(set)
        possible_objects_str = list(observed_objects.keys()) + list(domain.constants.keys())
        objects_and_consts = list(observed_objects.values()) + list(domain.constants.values())
        for predicate in domain.predicates.values():
            predicate_name = predicate.name
            signature_permutations = choose_objects_subset(possible_objects_str, len(predicate.signature))
            for signature_permutation in signature_permutations:
                grounded_signature = {
                    object_name: objects_and_consts[possible_objects_str.index(object_name)].type for object_name in signature_permutation
                }
                if not self._validate_type_matching(grounded_signature, predicate):
                    continue

                matching_grounded_type_hierarchy_signature = {
                    parameter_name: objects_and_consts[possible_objects_str.index(object_name)].type
                    for object_name, parameter_name in zip(grounded_signature, predicate.signature)
                }
                grounded_predicate = GroundedPredicate(
                    name=predicate_name,
                    signature=matching_grounded_type_hierarchy_signature,
                    object_mapping={parameter_name: object_name for object_name, parameter_name in zip(grounded_signature, predicate.signature)},
                )
                vocabulary[predicate.untyped_representation].add(grounded_predicate)

        return vocabulary

    def create_lifted_functions_vocabulary(
        self, domain: Union[LearnerDomain, Domain], possible_parameters: Dict[str, PDDLType], must_be_parameter: Optional[str] = None
    ) -> Dict[str, PDDLFunction]:
        """Create a function vocabulary from a domain containing action signatures and function definitions.

        :param domain: the domain containing the functions and the action signatures.
        :param possible_parameters: the parameters to use to create the vocabulary from.
        :param must_be_parameter: the parameter that must be in the function signature.
        :return: the vocabulary of functions.
        """
        self.logger.debug(f"Creating a function vocabulary from {possible_parameters}")
        vocabulary = {}
        possible_parameters_names = list(possible_parameters.keys()) + list(domain.constants.keys())
        parameter_types = list(possible_parameters.values()) + [const.type for const in domain.constants.values()]
        for predicate in domain.functions.values():
            function_name = predicate.name
            signature_permutations = choose_objects_subset(possible_parameters_names, len(predicate.signature))
            for signature_permutation in signature_permutations:
                bounded_lifted_signature = {
                    param_name: parameter_types[possible_parameters_names.index(param_name)] for param_name in signature_permutation
                }

                if not self._validate_type_matching(bounded_lifted_signature, predicate):
                    continue

                if must_be_parameter and must_be_parameter not in bounded_lifted_signature:
                    continue

                lifted_function = PDDLFunction(name=function_name, signature=bounded_lifted_signature)
                vocabulary[lifted_function.untyped_representation] = lifted_function

        self.logger.debug(f"Created a lifted function vocabulary of size {len(vocabulary)}")
        return vocabulary

    def create_lifted_vocabulary(
        self, domain: Union[LearnerDomain, Domain], possible_parameters: Dict[str, PDDLType], must_be_parameter: Optional[str] = None
    ) -> Set[Predicate]:
        """Create a vocabulary of random combinations of parameters that match the predicates.

        :param domain: the domain containing the predicates and the action signatures.
        :param possible_parameters: the parameters to use to create the vocabulary from.
        :param must_be_parameter: if not None, the vocabulary will only contain predicates that have this parameter.
        :return: list containing all the predicates with the different combinations of parameters.
        """
        self.logger.debug(f"Creating predicates vocabulary from {possible_parameters}")
        vocabulary = set()
        possible_parameters_names = list(possible_parameters.keys()) + list(domain.constants.keys())
        parameter_types = list(possible_parameters.values()) + [const.type for const in domain.constants.values()]
        for predicate in domain.predicates.values():
            predicate_name = predicate.name
            signature_permutations = choose_objects_subset(possible_parameters_names, len(predicate.signature))
            for signature_permutation in signature_permutations:
                bounded_lifted_signature = {
                    param_name: parameter_types[possible_parameters_names.index(param_name)] for param_name in signature_permutation
                }

                if not self._validate_type_matching(bounded_lifted_signature, predicate):
                    continue

                if must_be_parameter and must_be_parameter not in bounded_lifted_signature:
                    continue

                lifted_predicate = Predicate(name=predicate_name, signature=bounded_lifted_signature)
                negative_lifted_predicate = lifted_predicate.copy()
                negative_lifted_predicate.is_positive = False
                vocabulary.update({lifted_predicate, negative_lifted_predicate})

        self.logger.debug(f"Created vocabulary of size {len(vocabulary)}")
        return vocabulary

    def create_grounded_actions_vocabulary(self, domain: Union[LearnerDomain, Domain], observed_objects: Dict[str, PDDLObject]) -> Set[ActionCall]:
        """"Create a vocabulary of random combinations of the actions parameters and objects.

        :param domain: the domain containing the actions and the action signatures.
        :param observed_objects: the objects that were observed in the trajectory.
        :return: list containing all the actions with the different combinations of parameters.
        """
        vocabulary = set()
        objects_and_consts = list(observed_objects.values()) + list(domain.constants.values())
        for action_name, action in domain.actions.items():
            possible_objects_for_parameters = {
                parameter_name: [obj for obj in objects_and_consts if obj.type.is_sub_type(parameter_type)]
                for parameter_name, parameter_type in action.signature.items()
            }
            signature_options = list(itertools.product(*possible_objects_for_parameters.values()))
            for signature_option in signature_options:
                grounded_action = ActionCall(name=action_name, grounded_parameters=[obj.name for obj in signature_option])
                self.logger.debug(f"Created grounded action {str(grounded_action)}")
                vocabulary.add(grounded_action)

        return vocabulary



"""Module responsible for running the Expressive Numeric Heuristic Planner (ENHSP)."""
import logging
import os
import signal
import subprocess
import sys
from pathlib import Path
from typing import Dict

from jdk4py import JAVA

ENHSP_FILE_PATH = os.environ["ENHSP_FILE_PATH"]
MAX_RUNNING_TIME = 5  # seconds

TIMEOUT_ERROR_CODE = b"Timeout has been reached"
PROBLEM_SOLVED = b"Problem Solved"
NO_SOLUTION_FOR_PROBLEM = b"Problem Detected as Unsolvable"
OTHER_NO_SOLUTION_TYPE = b"Problem unsolvable"
GOAL_NOT_REACHABLE = b"Goal is not reachable"
ERROR_TYPE_NO_SOLUTION = b'"this.plan" is null'


class ENHSPSolver:
    """Class designated to use to activate the metric-FF solver on the cluster and parse its result."""

    logger: logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def _run_enhsp_process(
        self, run_command: str, problem_file_path: Path, solving_stats: Dict[str, str], solving_timeout: int = MAX_RUNNING_TIME
    ) -> bool:
        """Runs the ENHSP process and monitors its execution time.

        :param run_command: the command to run the ENHSP process.
        :param problem_file_path: the path to the problem file.
        :param solving_stats: the statistics of the solving process.
        :return: whether the process finished successfully.
        """
        self.logger.info(f"Starting to run ENHSP process for the problem - {problem_file_path.stem}")
        process = subprocess.Popen(run_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            process.wait(timeout=solving_timeout)

        except subprocess.TimeoutExpired:
            self.logger.warning(f"ENHSP did not finish after {solving_timeout} secs while trying to solve - {problem_file_path.stem}")
            solving_stats[problem_file_path.stem] = "timeout"
            os.kill(process.pid, signal.SIGTERM)
            os.system(f"pkill -f {ENHSP_FILE_PATH}")
            return True

        if process.returncode is None:
            solving_stats[problem_file_path.stem] = "timeout"
            self.logger.warning(f"ENHSP did not finish in time so was killed while trying to solve - {problem_file_path.stem}")
            return True

        self.logger.info("ENHSP finished its execution!")
        stdout = process.stdout.read()
        stderr = process.stderr.read()

        if PROBLEM_SOLVED in stdout:
            self.logger.info(f"Solver succeeded in solving problem - {problem_file_path.stem}")
            solving_stats[problem_file_path.stem] = "ok"
            return True

        elif NO_SOLUTION_FOR_PROBLEM in stdout or OTHER_NO_SOLUTION_TYPE in stdout or ERROR_TYPE_NO_SOLUTION in stderr:
            self.logger.warning(f"Solver could not solve problem - {problem_file_path.stem}")
            solving_stats[problem_file_path.stem] = "no_solution"
            return True

        elif GOAL_NOT_REACHABLE in stderr:
            self.logger.warning("Solver declared goal unreachable!")
            solving_stats[problem_file_path.stem] = "no_solution"
            return True

        else:
            self.logger.critical(f"While solving problem {problem_file_path.stem} encountered unknown error! " f"STDOUT - {stdout}")
            self.logger.critical(f"While solving problem {problem_file_path.stem} encountered unknown error! " f"STDERR - {stderr}")
            solving_stats[problem_file_path.stem] = "solver_error"
            return False

    def solve_problem(
        self,
        domain_file_path: Path,
        problem_file_path: Path,
        problems_directory_path: Path,
        solving_stats: Dict[str, str],
        solving_timeout: int,
        tolerance: float,
    ) -> None:
        """Solves a single problem using the ENHSP algorithm.

        :param domain_file_path: the path to the domain file.
        :param problem_file_path: the path to the problem file.
        :param problems_directory_path: the path to the problems' directory.
        :param solving_stats: the statistics of the solving process.
        :param solving_timeout: the timeout for the solver.
        :param tolerance: the numeric tolerance to use.
        """
        num_retries = 0
        self.logger.debug(f"Starting to work on solving problem - {problem_file_path.stem}")
        solution_path = problems_directory_path / f"{problem_file_path.stem}.solution"
        running_options = [
            "-o",
            str(domain_file_path.absolute()),
            "-f",
            str(problem_file_path.absolute()),
            "-planner",
            "sat-hmrphj",
            "-tolerance",
            f"{tolerance}",
            "-gro",
            "naive",
            "-sp",
            str(solution_path.absolute()),
        ]
        run_command = f"{str(JAVA)} -jar {ENHSP_FILE_PATH} {' '.join(running_options)}"
        solver_output_ok = self._run_enhsp_process(run_command, problem_file_path, solving_stats, solving_timeout)
        while not solver_output_ok and num_retries < 3:
            solver_output_ok = self._run_enhsp_process(run_command, problem_file_path, solving_stats, solving_timeout)
            num_retries += 1

    def execute_solver(
        self,
        problems_directory_path: Path,
        domain_file_path: Path,
        solving_timeout: int = MAX_RUNNING_TIME,
        tolerance: float = 0.1,
        problems_prefix: str = "pfile",
    ) -> Dict[str, str]:
        """Solves numeric and PDDL+ problems using the ENHSP algorithm, automatically outputs the solution into a file.

        :param problems_directory_path: the path to the problems directory.
        :param domain_file_path: the path to the domain file.
        :param solving_timeout: the timeout for the solver.
        :param tolerance: the numeric tolerance to use.
        :param problems_prefix: the prefix of the problems to solve.
        """
        solving_stats = {}
        self.logger.info("Starting to solve the input problems using ENHSP solver.")
        for problem_file_path in problems_directory_path.glob(f"{problems_prefix}*.pddl"):
            self.solve_problem(domain_file_path, problem_file_path, problems_directory_path, solving_stats, solving_timeout, tolerance)

        return solving_stats


if __name__ == "__main__":
    args = sys.argv
    logging.basicConfig(format="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.DEBUG)
    solver = ENHSPSolver()
    solver.solve_problem(
        domain_file_path=Path(args[1]),
        problem_file_path=Path(args[2]),
        problems_directory_path=Path(args[3]),
        solving_stats={},
        solving_timeout=int(args[4]),
        tolerance=float(args[5]),
    )



"""Module to manage the action model learning statistics."""
import csv
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Domain, Observation, MultiAgentObservation, MultiAgentComponent, Predicate

from sam_learning.core import LearnerDomain, VocabularyCreator
from statistics.discrete_precision_recall_calculator import PrecisionRecallCalculator
from utilities import LearningAlgorithmType, NegativePreconditionPolicy

LEARNED_ACTIONS_STATS_COLUMNS = [
    "learning_algorithm",
    "policy",
    "domain_name",
    "num_trajectories",
    "learning_time",
    "num_trajectory_triplets",
    "total_number_of_actions",
    "learned_action_name",
    "num_triplets_action_appeared",
    "num_pb_literals",
    "num_pb_functions",
    "learned_discrete_preconditions",
    "num_learned_positive_preconditions",
    "num_learned_negative_preconditions",
    "num_learned_add_effects",
    "num_learned_delete_effects",
    "num_ground_truth_positive_preconditions",
    "num_ground_truth_negative_preconditions",
    "num_ground_truth_add_effects",
    "num_ground_truth_delete_effects",
    "learned_discrete_effects",
    "num_discrete_effects",
    "ground_truth_preconditions",
    "ground_truth_effects",
    "preconditions_precision",
    "effects_precision",
    "preconditions_recall",
    "effects_recall",
    "action_precision",
    "action_recall",
    "f1_score",
]
UNSAFE_LEARNING_ALGORITHMS = [
    LearningAlgorithmType.plan_miner,
]


class LearningStatisticsManager:
    """Class that manages the statistics about the action's learning properties gathered from the learning process."""

    model_domain: Domain
    learning_algorithm: LearningAlgorithmType
    working_directory_path: Path
    results_dir_path: Path
    action_learning_stats: List[Dict[str, Any]]
    numeric_learning_stats: List[Dict[str, Any]]
    merged_numeric_stats: List[Dict[str, Any]]
    merged_action_stats: List[Dict[str, Any]]

    def __init__(self, working_directory_path: Path, domain_path: Path, learning_algorithm: LearningAlgorithmType):
        self.working_directory_path = working_directory_path
        self.model_domain = DomainParser(domain_path=domain_path, partial_parsing=False).parse_domain()
        self.learning_algorithm = learning_algorithm
        self.action_learning_stats = []
        self.numeric_learning_stats = []
        self.merged_action_stats = []
        self.merged_numeric_stats = []
        self.results_dir_path = self.working_directory_path / "results_directory"
        self.vocabulary_creator = VocabularyCreator()

    @staticmethod
    def _update_action_appearances(used_observations: List[Union[Observation, MultiAgentObservation]]) -> Counter:
        """Updates the number of times each action appeared in the input trajectories.

        :param used_observations: the observations used to learn the actions.
        :return: Counter object mapping the action name to the number of appearances.
        """
        action_appearance_counter = Counter()
        for observation in used_observations:
            for component in observation.components:
                if isinstance(component, MultiAgentComponent):
                    for action in component.grounded_joint_action.actions:
                        action_appearance_counter[action.name] += 1
                else:
                    action_appearance_counter[component.grounded_action_call.name] += 1
        return action_appearance_counter

    def _extract_all_preconditions(self, action_name, learned_domain):
        learned_preconditions = [p.untyped_representation for _, p in learned_domain.actions[action_name].preconditions if isinstance(p, Predicate)]
        ground_truth_preconditions = [
            p.untyped_representation for _, p in self.model_domain.actions[action_name].preconditions if isinstance(p, Predicate)
        ]
        return ground_truth_preconditions, learned_preconditions

    def create_results_directory(self) -> None:
        """Creates the results' directory that contains the learning results."""
        self.results_dir_path.mkdir(exist_ok=True)

    def _add_default_statistic_fields(
        self,
        action_name: str,
        learning_report: Optional[Dict[str, str]],
        policy: NegativePreconditionPolicy,
        used_observations: List[Union[Observation, MultiAgentObservation]],
    ) -> Dict[str, Any]:
        """Adds the default fields to the action statistics (those not inferred from the learning process).

        :param action_name: the name of the action.
        :param learning_report: the report on the status of the learned actions, whether they were safe to learn or not.
        :param policy: the policy used to remove the negative preconditions from the action.
        :param used_observations: the observations that were used to learn the action.
        :return: the dictionary containing the default fields.
        """
        learning_time = float(learning_report["learning_time"])
        action_appearance_counter = self._update_action_appearances(used_observations)
        num_triplets = sum([len(observation.components) for observation in used_observations])
        ground_truth_preconditions = [p for _, p in self.model_domain.actions[action_name].preconditions if isinstance(p, Predicate)]
        ground_truth_effects = [p for p in self.model_domain.actions[action_name].discrete_effects]
        action_pbls_vocabulary = self.vocabulary_creator.create_lifted_vocabulary(
            domain=self.model_domain, possible_parameters=self.model_domain.actions[action_name].signature
        )
        action_pb_function_vocabulary = self.vocabulary_creator.create_lifted_functions_vocabulary(
            domain=self.model_domain, possible_parameters=self.model_domain.actions[action_name].signature
        )
        return {
            "learning_algorithm": self.learning_algorithm.name,
            "policy": policy.name,
            "domain_name": self.model_domain.name,
            "num_trajectories": len(used_observations),
            "learning_time": learning_time,
            "num_trajectory_triplets": num_triplets,
            "total_number_of_actions": len(self.model_domain.actions),
            "learned_action_name": action_name,
            "num_triplets_action_appeared": action_appearance_counter[action_name],
            "num_pb_literals": len(action_pbls_vocabulary),
            "num_pb_functions": len(action_pb_function_vocabulary),
            "num_ground_truth_positive_preconditions": len([p for p in ground_truth_preconditions if p.is_positive]),
            "num_ground_truth_negative_preconditions": len([p for p in ground_truth_preconditions if not p.is_positive]),
            "num_ground_truth_add_effects": len([p for p in ground_truth_effects if p.is_positive]),
            "num_ground_truth_delete_effects": len([p for p in ground_truth_effects if not p.is_positive]),
            "ground_truth_preconditions": [p.untyped_representation for p in ground_truth_preconditions],
            "ground_truth_effects": [p.untyped_representation for p in ground_truth_effects],
        }

    def add_to_action_stats(
        self,
        used_observations: List[Union[Observation, MultiAgentObservation]],
        learned_domain: LearnerDomain,
        learning_report: Optional[Dict[str, str]] = None,
        policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
    ) -> None:
        """Add the action data to the statistics.

        :param used_observations: the observations that were used to learn the action.
        :param learned_domain: the domain that was learned from the action model learning algorithm.
        :param learning_report: the report on the status of the learned actions, whether they were safe to learn or not.
        :param policy: the policy used to remove the negative preconditions from the action.
        """
        precision_recall_calculator = PrecisionRecallCalculator()
        for action_name, action_data in self.model_domain.actions.items():
            if action_name not in learned_domain.actions:
                action_stats = {
                    **self._add_default_statistic_fields(action_name, learning_report, policy, used_observations),
                    "learned_discrete_preconditions": [],
                    "learned_discrete_effects": [],
                    "num_learned_positive_preconditions": 0,
                    "num_learned_negative_preconditions": 0,
                    "num_learned_add_effects": 0,
                    "num_learned_delete_effects": 0,
                    "preconditions_precision": 0,
                    "effects_precision": 1,
                    "preconditions_recall": 1,
                    "effects_recall": 0,
                    "action_precision": 0,
                    "action_recall": 1,
                    "f1_score": 0,
                }
                self.action_learning_stats.append(action_stats)
                continue

            ground_truth_preconditions, learned_preconditions = self._extract_all_preconditions(action_name, learned_domain)
            precision_recall_calculator.add_action_data(learned_domain.actions[action_name], self.model_domain.actions[action_name])
            discrete_preconditions = [p for _, p in learned_domain.actions[action_name].preconditions if isinstance(p, Predicate)]
            learned_discrete_effects = [p.untyped_representation for p in learned_domain.actions[action_name].discrete_effects]
            action_stats = {
                **self._add_default_statistic_fields(action_name, learning_report, policy, used_observations),
                "learned_discrete_preconditions": learned_preconditions,
                "learned_discrete_effects": learned_discrete_effects,
                "num_learned_positive_preconditions": len([p for p in discrete_preconditions if p.is_positive]),
                "num_learned_negative_preconditions": len([p for p in discrete_preconditions if not p.is_positive]),
                "num_learned_add_effects": len([p for p in learned_domain.actions[action_name].discrete_effects if p.is_positive]),
                "num_learned_delete_effects": len([p for p in learned_domain.actions[action_name].discrete_effects if not p.is_positive]),
                **precision_recall_calculator.export_action_syntactic_statistics(action_name),
            }
            self.action_learning_stats.append(action_stats)

    def _export_statistics_data(
        self,
        fold_number: int,
        iteration_num: Optional[int] = None,
        columns: List[str] = [],
        action_data_to_export: List[Dict[str, Any]] = {},
        stats_data_file_name: str = "",
    ) -> None:
        """Exports statistics to a report CSV file.

        :param fold_number: the number of the currently running fold.
        :param iteration_num: the number of the current iteration.
        :param columns: the column names in the CSV report file.
        :param action_data_to_export: the data to export to the report.
        :param stats_data_file_name: the name of the statistics to publish in the report file.
        """
        iterations_suffix = f"_{iteration_num}" if iteration_num is not None else ""
        statistics_path = (
            self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}"
            f"_{stats_data_file_name}_{fold_number}{iterations_suffix}.csv"
        )
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=columns)
            stats_writer.writeheader()
            for data_line in action_data_to_export:
                stats_writer.writerow(data_line)

    def export_action_learning_statistics(self, fold_number: int, iteration_num: Optional[int] = None) -> None:
        """Export the statistics collected about the actions.

        :param fold_number: the number of the currently running fold.
        :param iteration_num: the number of the current iteration.
        """
        self._export_statistics_data(fold_number, iteration_num, LEARNED_ACTIONS_STATS_COLUMNS, self.action_learning_stats, "action_stats_fold")

    def export_all_folds_action_stats(self) -> None:
        """Export the statistics collected about the actions."""
        output_path = self.results_dir_path / f"all_folds_action_learning_stats.csv"
        with open(output_path, "wt", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=LEARNED_ACTIONS_STATS_COLUMNS)
            writer.writeheader()
            writer.writerows(self.merged_action_stats)

    def clear_statistics(self) -> None:
        """Clears the statistics so that each fold will have no relation to its predecessors."""
        self.merged_numeric_stats.extend(self.numeric_learning_stats)
        self.merged_action_stats.extend(self.action_learning_stats)
        self.numeric_learning_stats.clear()
        self.action_learning_stats.clear()

    def write_complete_joint_statistics(self) -> None:
        """Writes a statistics file containing all the folds combined data."""
        output_path = self.results_dir_path / f"{self.learning_algorithm.name}_all_folds_numeric_learning_stats.csv"
        with open(output_path, "wt", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=LEARNED_ACTIONS_STATS_COLUMNS)
            writer.writeheader()
            writer.writerows(self.merged_numeric_stats)



"""Extension to SAM Learning that can learn numeric state variables."""

from typing import List, Dict, Tuple, Optional

from pddl_plus_parser.models import Observation, ActionCall, State, Domain

from sam_learning.core import LearnerDomain, NumericFunctionMatcher, NotSafeActionError, LearnerAction
from sam_learning.core.learner_domain import DISJUNCTIVE_PRECONDITIONS_REQ
from sam_learning.core.baseline_algorithms_version.naive_numeric_fluent_learner_algorithm import NaiveNumericFluentStateStorage
from sam_learning.core.baseline_algorithms_version.naive_polynomial_fluents_learning_algorithm import NaivePolynomialFluentsLearningAlgorithm
from sam_learning.learners.sam_learning import SAMLearner


class NaiveNumericSAMLearner(SAMLearner):
    """The Extension of SAM that is able to learn numeric state variables."""

    storage: Dict[str, NaiveNumericFluentStateStorage]
    function_matcher: NumericFunctionMatcher
    relevant_fluents: Dict[str, List[str]]

    def __init__(
        self,
        partial_domain: Domain,
        relevant_fluents: Optional[Dict[str, List[str]]] = None,
        effects_fluent_map: Optional[Dict[str, List[str]]] = None,
        **kwargs,
    ):
        super().__init__(partial_domain)
        self.storage = {}
        self.function_matcher = NumericFunctionMatcher(partial_domain)
        self.relevant_fluents = relevant_fluents
        self.effects_fluent_map = effects_fluent_map

    def _construct_safe_numeric_preconditions(self, action: LearnerAction) -> None:
        """Constructs the safe preconditions for the input action.

        :param action: the action that the preconditions are constructed for.
        """
        action_name = action.name
        if self.relevant_fluents is None:
            learned_numeric_preconditions = self.storage[action_name].construct_safe_linear_inequalities()

        elif len(self.relevant_fluents[action_name]) == 0:
            self.logger.debug(f"The action {action_name} has no numeric preconditions.")
            return

        else:
            learned_numeric_preconditions = self.storage[action_name].construct_safe_linear_inequalities(self.relevant_fluents[action_name])

        if learned_numeric_preconditions.binary_operator == "and":
            self.logger.debug("The learned preconditions are a conjunction. Adding the internal numeric conditions.")
            for condition in learned_numeric_preconditions.operands:
                action.preconditions.add_condition(condition)

            return

        self.logger.debug("The learned preconditions are not a conjunction. Adding them as a separate condition.")
        action.preconditions.add_condition(learned_numeric_preconditions)
        self.partial_domain.requirements.append(DISJUNCTIVE_PRECONDITIONS_REQ)

    def _construct_safe_numeric_effects(self, action: LearnerAction) -> None:
        """Constructs the safe numeric effects for the input action.

        :param action: the action that its effects are constructed for.
        """
        effects, numeric_preconditions, learned_perfectly = self.storage[action.name].construct_assignment_equations(
            relevant_fluents=self.relevant_fluents[action.name] if self.relevant_fluents is not None else None
        )
        if effects is not None and len(effects) > 0:
            action.numeric_effects = effects

        if self.relevant_fluents is None:
            self.logger.debug(f"No feature selection applied, using the numeric preconditions as is.")
            return

        self.logger.info(f"The effect of action - {action.name} were learned perfectly.")
        if numeric_preconditions is not None:
            for cond in numeric_preconditions.operands:
                if cond in action.preconditions.root:
                    continue

                action.preconditions.add_condition(cond)

        return

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Adds a new action to the learned domain.

        :param grounded_action: the grounded action that was executed according to the observation.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
        """
        super().add_new_action(grounded_action, previous_state, next_state)
        self.logger.debug(f"Creating the new storage for the action - {grounded_action.name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        self.storage[grounded_action.name] = NaiveNumericFluentStateStorage(grounded_action.name, self.partial_domain.functions)
        self.storage[grounded_action.name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[grounded_action.name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done creating the numeric state variable storage for the action - {grounded_action.name}")

    def update_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Updates the action's data according to the new input observed triplet.

        :param grounded_action: the grounded action that was observed.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.name
        super().update_action(grounded_action, previous_state, next_state)
        self.logger.debug(f"Adding the numeric state variables to the numeric storage of action - {action_name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        self.storage[action_name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[action_name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done updating the numeric state variable storage for the action - {grounded_action.name}")

    def learn_action_model(self, observations: List[Observation]) -> Tuple[LearnerDomain, Dict[str, str]]:
        """Learn the SAFE action model from the input observations.

        :param observations: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned and the metadata about the learning.
        """
        self.logger.info("Starting to learn the action model!")
        super().start_measure_learning_time()
        allowed_actions = {}
        learning_metadata = {}
        super().deduce_initial_inequality_preconditions()
        for observation in observations:
            self.current_trajectory_objects = observation.grounded_objects
            for component in observation.components:
                if not super().are_states_different(component.previous_state, component.next_state):
                    continue

                self.handle_single_trajectory_component(component)

        for action_name, action in self.partial_domain.actions.items():
            if action_name not in self.storage:
                self.logger.debug(f"The action - {action_name} has not been observed in the trajectories!")
                continue

            self.storage[action_name].filter_out_inconsistent_state_variables()
            try:
                self._construct_safe_numeric_preconditions(action)
                self._construct_safe_numeric_effects(action)
                allowed_actions[action_name] = action
                learning_metadata[action_name] = "OK"
                self.logger.info(f"Done learning the action - {action_name}!")

            except NotSafeActionError as e:
                self.logger.debug(f"The action - {e.action_name} is not safe for execution, reason - {e.reason}")
                learning_metadata[action_name] = e.solution_type.name

        self.partial_domain.actions = allowed_actions

        super().end_measure_learning_time()
        learning_metadata["learning_time"] = str(self.learning_end_time - self.learning_start_time)
        return self.partial_domain, learning_metadata


class NaivePolynomialSAMLearning(NaiveNumericSAMLearner):
    """The Extension of SAM that is able to learn polynomial state variables."""

    storage: Dict[str, NaivePolynomialFluentsLearningAlgorithm]
    polynom_degree: int

    def __init__(self, partial_domain: Domain, relevant_fluents: Optional[Dict[str, List[str]]] = None, polynomial_degree: int = 1, **kwargs):
        super().__init__(partial_domain, relevant_fluents, **kwargs)
        self.polynom_degree = polynomial_degree

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Adds a new action to the learned domain.

        :param grounded_action: the grounded action that was executed according to the observation.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        super().add_new_action(grounded_action, previous_state, next_state)
        self.logger.debug(f"Creating the new storage for the action - {grounded_action.name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, previous_state.state_fluents)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, next_state.state_fluents)
        self.storage[grounded_action.name] = NaivePolynomialFluentsLearningAlgorithm(
            grounded_action.name, self.polynom_degree, self.partial_domain.functions, is_verbose=True
        )
        self.storage[grounded_action.name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[grounded_action.name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done creating the numeric state variable storage for the action - {grounded_action.name}")



# Add current project to sys path
import contextlib
import json
import logging
import os
import sys

import pandas as pd
from unified_planning.engines import ValidationResultStatus, PlanGenerationResultStatus

parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)

import copy
import warnings

from alive_progress import alive_bar

from amlgym.util.SimpleDomainReader import SimpleDomainReader
import numpy as np
import os
import re
import shutil
from typing import List, Dict, Set

import unified_planning.model
from tarski.grounding import LPGroundingStrategy
from unified_planning.io import PDDLReader, PDDLWriter
from tarski.io import PDDLReader as tarskiPDDLReader
from unified_planning.plans import ActionInstance
from unified_planning.shortcuts import SequentialSimulator, PlanValidator, OneshotPlanner

# Disable printing of planning engine credits to avoid overloading stdout
unified_planning.shortcuts.get_environment().credits_stream = None


def syntactic_eval(model_learned, model_reference):

    eval_model = SimpleDomainReader(input_file=model_learned)
    model_reference = SimpleDomainReader(input_file=model_reference)

    for op_gt, op_eval in zip(model_reference.operators, eval_model.operators):
        op_gt.operator_name = op_gt.operator_name.replace('_', '-')
        op_eval.operator_name = op_eval.operator_name.replace('_', '-')

    # Sort operators list
    model_reference.operators = sorted(model_reference.operators, key=lambda x: x.operator_name, reverse=True)
    eval_operatos_name = [op.operator_name for op in eval_model.operators]
    sorted_eval_operators = []
    for op in model_reference.operators:
        if op.operator_name in eval_operatos_name:
            operator = next((o for o in eval_model.operators if o.operator_name == op.operator_name), None)
            sorted_eval_operators.append(operator)
        else:
            new_operator = copy.deepcopy(op)
            new_operator.precs_pos = []
            new_operator.precs_neg = []
            new_operator.eff_pos = []
            new_operator.eff_neg = []
            sorted_eval_operators.append(new_operator)
    eval_model.operators = sorted_eval_operators

    assert bool(np.all([gt_op.operator_name == learned_op.operator_name for gt_op, learned_op in
                        zip(model_reference.operators, eval_model.operators)]))

    all_pre_pos_precision = []
    all_pre_neg_precision = []
    all_eff_pos_precision = []
    all_eff_neg_precision = []
    all_pre_pos_recall = []
    all_pre_neg_recall = []
    all_eff_pos_recall = []
    all_eff_neg_recall = []
    overall_recall = []
    overall_precision = []

    for gt_op, learned_op in zip(model_reference.operators, eval_model.operators):
        fn_pre_pos = len(set([p for p in gt_op.precs_pos if p not in learned_op.precs_pos]))
        fp_pre_pos = len(set([p for p in learned_op.precs_pos if p not in gt_op.precs_pos]))
        tp_pre_pos = len(set([p for p in learned_op.precs_pos if p in gt_op.precs_pos]))

        fn_pre_neg = len(set([p for p in gt_op.precs_neg if p not in learned_op.precs_neg]))
        fp_pre_neg = len(set([p for p in learned_op.precs_neg if p not in gt_op.precs_neg]))
        tp_pre_neg = len(set([p for p in learned_op.precs_neg if p in gt_op.precs_neg]))

        fn_eff_pos = len(set([p for p in gt_op.eff_pos if p not in learned_op.eff_pos]))
        fp_eff_pos = len(set([p for p in learned_op.eff_pos if p not in gt_op.eff_pos]))
        tp_eff_pos = len(set([p for p in learned_op.eff_pos if p in gt_op.eff_pos]))

        fn_eff_neg = len(set([p for p in gt_op.eff_neg if p not in learned_op.eff_neg]))
        fp_eff_neg = len(set([p for p in learned_op.eff_neg if p not in gt_op.eff_neg]))
        tp_eff_neg = len(set([p for p in learned_op.eff_neg if p in gt_op.eff_neg]))

        all_tp = tp_pre_pos + tp_eff_pos + tp_eff_neg + tp_pre_neg
        all_fp = fp_pre_pos + fp_pre_neg + fp_eff_pos + fp_eff_neg
        all_fn = fn_pre_pos + fn_pre_neg + fn_eff_pos + fn_eff_neg

        pre_pos_recall = pre_pos_precision = None
        pre_neg_recall = pre_neg_precision = None
        eff_pos_recall = eff_pos_precision = None
        eff_neg_recall = eff_neg_precision = None

        if tp_pre_pos + fp_pre_pos > 0:
            pre_pos_precision = tp_pre_pos / (tp_pre_pos + fp_pre_pos)
        else:
            warnings.warn('No positive precondition has been learned '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Precision of positive preconditions for '
                          f'{gt_op.operator_name} evaluated as 1.')
            pre_pos_precision = 1.

        if tp_pre_neg + fp_pre_neg > 0:
            pre_neg_precision = tp_pre_neg / (tp_pre_neg + fp_pre_neg)
        else:
            warnings.warn('No negative precondition has been learned '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}.'
                          f' Precision of negative preconditions for '
                          f'{gt_op.operator_name} evaluated as 1.')
            pre_neg_precision = 1.

        if tp_eff_pos + fp_eff_pos > 0:
            eff_pos_precision = tp_eff_pos / (tp_eff_pos + fp_eff_pos)
        else:
            warnings.warn('No positive effect has been learned '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Precision of positive effects for '
                          f'{gt_op.operator_name} evaluated as 1.')
            eff_pos_precision = 1.

        if tp_eff_neg + fp_eff_neg > 0:
            eff_neg_precision = tp_eff_neg / (tp_eff_neg + fp_eff_neg)
        else:
            warnings.warn('No negative effect has been learned '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Precision of negative effects for '
                          f'{gt_op.operator_name} evaluated as 1.')
            eff_neg_precision = 1.

        if tp_pre_pos + fn_pre_pos > 0:
            pre_pos_recall = tp_pre_pos / (tp_pre_pos + fn_pre_pos)
        else:
            assert len(gt_op.precs_pos) == 0
            warnings.warn('No positive precondition exists '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Recall of positive preconditions for '
                          f'{gt_op.operator_name} evaluated as 1.')
            pre_pos_recall = 1.

        if tp_pre_neg + fn_pre_neg > 0:
            pre_neg_recall = tp_pre_neg / (tp_pre_neg + fn_pre_neg)
        else:
            assert len(gt_op.precs_neg) == 0
            warnings.warn('No negative precondition exists '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Recall of negative preconditions for '
                          f'{gt_op.operator_name} evaluated as 1.')
            pre_neg_recall = 1.

        if tp_eff_pos + fn_eff_pos > 0:
            eff_pos_recall = tp_eff_pos / (tp_eff_pos + fn_eff_pos)
        else:
            assert len(gt_op.eff_pos) == 0
            warnings.warn('No positive effect exists '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Recall of positive effects for '
                          f'{gt_op.operator_name} evaluated as 1.')
            eff_pos_recall = 1.

        if tp_eff_neg + fn_eff_neg > 0:
            eff_neg_recall = tp_eff_neg / (tp_eff_neg + fn_eff_neg)
        else:
            assert len(gt_op.eff_neg) == 0
            warnings.warn('No negative effect exists '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Recall of negative effects for '
                          f'{gt_op.operator_name} evaluated as 1.')
            eff_neg_recall = 1.

        if all_tp + all_fp > 0:
            overall_precision.append(all_tp / (all_tp + all_fp))
        else:
            warnings.warn('No positive predictions in terms of pos/neg precs/effs '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Overall precision '
                          f'{gt_op.operator_name} evaluated as 1.')
            overall_precision.append(1.0)

        if all_tp + all_fn > 0:
            overall_recall.append(all_tp / (all_tp + all_fn))
        else:
            warnings.warn('No negative predictions in terms of pos/neg precs/effs '
                          f'for operator {gt_op.operator_name} of domain'
                          f' {model_reference.input_file}. '
                          f'Overall precision '
                          f'{gt_op.operator_name} evaluated as 1.')
            overall_recall.append(1.0)

        all_pre_pos_precision.append(pre_pos_precision)
        all_pre_neg_precision.append(pre_neg_precision)
        all_eff_pos_precision.append(eff_pos_precision)
        all_eff_neg_precision.append(eff_neg_precision)
        all_pre_pos_recall.append(pre_pos_recall)
        all_pre_neg_recall.append(pre_neg_recall)
        all_eff_pos_recall.append(eff_pos_recall)
        all_eff_neg_recall.append(eff_neg_recall)

    return {'syntactic':
                {
                    'precs_pos_recall': "{0:.2f}".format(np.mean(all_pre_pos_recall)),
                    'precs_neg_recall': "{0:.2f}".format(np.mean(all_pre_neg_recall)),
                    'pos_recall': "{0:.2f}".format(np.mean(all_eff_pos_recall)),
                    'neg_recall': "{0:.2f}".format(np.mean(all_eff_neg_recall)),
                    'precs_pos_precision': "{0:.2f}".format(np.mean(all_pre_pos_precision)),
                    'precs_neg_precision': "{0:.2f}".format(np.mean(all_pre_neg_precision)),
                    'pos_precision': "{0:.2f}".format(np.mean(all_eff_pos_precision)),
                    'neg_precision': "{0:.2f}".format(np.mean(all_eff_neg_precision)),
                    'overall_recall': "{0:.2f}".format(np.mean(overall_recall)),
                    'overall_precision': "{0:.2f}".format(np.mean(overall_precision))
                }
    }


# TODO: currently keep it as it is for efficiency
def predictive_eval(model_learned_path: str,
                    model_ref_path: str,
                    traj_problem_paths: Dict[str, str]) -> Dict[str, Dict[str, float] | float]:
    """
    Compute the action applicability (precision and recall) of a learned model w.r.t. a
     reference model by considering all states in the input trajectories

    :param model_learned_path: file path to learned pddl model
    :param model_ref_path: file path to reference pddl model
    :param traj_problem_paths: a dictionary where each key is a path to a pddl trajectory
        and the value a path to the associated pddl problem. This is required as input since
        the trajectory does not explicitly report the set of problem objects, which is
        required to evaluate action applicability by simulating actions.

    :return: action applicability precision and recall grouped by operator
    """

    # Inner helper function for computing all applicable actions in a state
    def get_applicable_actions(simulator: SequentialSimulator,
                               state: unified_planning.model.State,
                               problem,
                               # for efficiency pass the ground actions
                               ground_actions: Dict[str, any]) -> Dict[str, Set[str]]:
        applicable_actions = {op_name:
                                  {objs for objs in all_objs
                                   if simulator._is_applicable(state,
                                                               problem.action(op_name),
                                                               [problem.object(o.lower()) for o in objs])}
                              for op_name, all_objs in ground_actions.items()
                              }
        return applicable_actions

    # Get the reference domain operators
    reader = PDDLReader()
    upl_domain = PDDLReader().parse_problem(model_ref_path)
    operators = {
        a.name: {
            'applicability': {
                'tp': 0,  # true positives
                'fp': 0,  # false positives
                'fn': 0,  # false negatives
                'precision': 0.,
                'recall': 0.,
            },
            'consequentiality': {
                'tp': 0,  # true positives
                'fp': 0,  # false positives
                'fn': 0,  # false negatives
                'precision': 0.,
                'recall': 0.,
            }
        }
        for a in upl_domain.actions}

    # Iterate over each trajectory
    with alive_bar(len(traj_problem_paths),
                   title=f'Evaluating applicability for domain {model_learned_path}',
                   length=20,
                   bar='smooth') as bar:

        for i, t in enumerate(sorted(traj_problem_paths.keys(),
                                     key=lambda x: int(x.split('/')[-1].split('_')[0]))):

            # Read trajectory data
            with open(t, 'r') as f:
                traj_data = f.readlines()
                states = [s for s in traj_data if s.startswith("(:state ")]
                actions = [s for s in traj_data if s.startswith("(:action ")]
                plan_str = '\n'.join([a.replace('(:action ', '').strip()[:-1] for a in actions])

            # Update problem initial state with the initial state of the trajectory. Notice they may be different
            # when the trajectory is obtained by sampling a subtrajectory from the originally generated one.
            problem_file = traj_problem_paths[t]
            shutil.copy(problem_file, 'tmp.pddl')
            with open('tmp.pddl', 'r') as f:
                problem_str = f.read()
            new_atoms = re.findall(r"\([^()]*\)", states[0])
            # Create the new init block
            new_init = "(:init\n" + "\n".join(new_atoms) + "\n)"
            # Replace the init block
            updated_problem = re.sub(r"\(:init[\s\S]*?\)(?=\s*\(:goal)", new_init, problem_str, flags=re.MULTILINE)
            with open('tmp.pddl', 'w') as f:
                # TODO: open issue in unified-planning
                updated_problem = updated_problem.replace('(at_ ', '(at ')
                f.write(updated_problem)

            # Parse unified-planning problem
            problem = PDDLReader().parse_problem(model_ref_path, 'tmp.pddl')
            os.remove('tmp.pddl')

            # Parse unified-planning plan
            plan = reader.parse_plan_string(problem, plan_str)

            # Ground actions with tarski since unified-planning (1.2.0) grounder is inefficient
            tarski_reader = tarskiPDDLReader(raise_on_error=True)
            tarski_reader.parse_domain(model_ref_path)
            tarski_reader.parse_instance(problem_file)
            grounder = LPGroundingStrategy(tarski_reader.problem)
            ground_actions = grounder.ground_actions()

            # Instantiate the simulator
            with SequentialSimulator(problem=problem) as simulator_ref:

                # Get initial state
                current_state = simulator_ref.get_initial_state()
                lpos = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*true', str(current_state))
                lneg = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*false', str(current_state))
                lprev = set(lpos).union(set([f'not_{l}' for l in lneg]))

                # Get applicable actions in the initial state according to the reference model
                app_actions_ref = get_applicable_actions(simulator_ref, current_state, problem, ground_actions)

                # Parse the simulator state into a pddl problem file for checking action applicability
                # in the learned domain, which requires instantiating a `learned simulator`
                new_problem = problem.clone()
                for fluent in problem.initial_values:
                    value = current_state.get_value(fluent)
                    new_problem.set_initial_value(fluent, value)
                PDDLWriter(new_problem).write_problem("tmp.pddl")

                # TODO: open issue in unified-planning
                with open("tmp.pddl", 'r') as f:
                    data = f.read()
                with open("tmp.pddl", 'w') as f:
                    f.write(data.replace("(at_ ", "(at "))

                domain_learned = PDDLReader().parse_problem(model_learned_path, "tmp.pddl")
                simulator_learned = SequentialSimulator(problem=domain_learned)
                os.remove('tmp.pddl')
                app_actions_learned = get_applicable_actions(simulator_learned,
                                                             # current_state,
                                                             simulator_learned.get_initial_state(),
                                                             domain_learned,
                                                             ground_actions)

                for op in operators:
                    operators[op]['applicability']['tp'] += len(app_actions_learned[op].intersection(app_actions_ref[op]))
                    operators[op]['applicability']['fp'] += len(app_actions_learned[op] - (app_actions_ref[op]))
                    operators[op]['applicability']['fn'] += len(app_actions_ref[op] - (app_actions_learned[op]))

                # Simulate the plan
                # TODO: refactor this if there are no efficiency issues
                for a in plan.actions:

                    prev_problem = problem.clone()
                    for fluent in problem.initial_values:
                        value = current_state.get_value(fluent)
                        prev_problem.set_initial_value(fluent, value)
                    PDDLWriter(prev_problem).write_problem("tmp.pddl")
                    # TODO: open issue in unified-planning
                    with open("tmp.pddl", 'r') as f:
                        data = f.read()
                    with open("tmp.pddl", 'w') as f:
                        f.write(data.replace("(at_ ", "(at "))

                    domain_learned = PDDLReader().parse_problem(model_learned_path, "tmp.pddl")

                    # Must reinstantiate with `new_problem` because of unified-planning object references...
                    simulator_learned = SequentialSimulator(problem=domain_learned)
                    os.remove('tmp.pddl')
                    a_learn = ActionInstance(domain_learned.action(a.action.name),
                                             [domain_learned.object(str(o)) for o in a.actual_parameters])

                    applicable_learn = simulator_learned._is_applicable(simulator_learned.get_initial_state(),
                                                            domain_learned.action(a.action.name),
                                                            [domain_learned.object(str(o)) for o in a.actual_parameters])
                    if applicable_learn:
                        next_state_learned = simulator_learned.apply(simulator_learned.get_initial_state(), a_learn)  # init state equals the ref one
                        assert next_state_learned is not None

                        lpos = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*true', str(next_state_learned))
                        lneg = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*false', str(next_state_learned))
                        lnext_learned = set(lpos).union(set([f'not_{l}' for l in lneg]))

                    # Check action is applicable in the environment model
                    applicable_ref = simulator_ref._is_applicable(current_state,
                                                            problem.action(a.action.name),
                                                            [problem.object(str(o)) for o in a.actual_parameters])
                    assert applicable_ref, ('the action of a plan produced by the environment model '
                                            'must be applicable according to the environment model.')
                    if applicable_ref:
                        current_state = simulator_ref.apply(current_state, a)
                        lpos = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*true', str(current_state))
                        lneg = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*false', str(current_state))
                        lnext_ref = set(lpos).union(set([f'not_{l}' for l in lneg]))

                        if current_state is None:
                            raise Exception(f"Error in applying: {a}")

                    # Get applicable actions in the initial state according to the reference model
                    app_actions_ref = get_applicable_actions(simulator_ref,
                                                             current_state,
                                                             problem,
                                                             ground_actions)

                    # Parse the simulator state into a pddl problem file for checking action applicability
                    # in the learned domain, which requires instantiating a `learned simulator`
                    new_problem = problem.clone()
                    for fluent in problem.initial_values:
                        value = current_state.get_value(fluent)
                        new_problem.set_initial_value(fluent, value)
                    PDDLWriter(new_problem).write_problem("tmp.pddl")
                    # TODO: open issue in unified-planning
                    with open("tmp.pddl", 'r') as f:
                        data = f.read()
                    with open("tmp.pddl", 'w') as f:
                        f.write(data.replace("(at_ ", "(at "))

                    domain_learned = PDDLReader().parse_problem(model_learned_path, "tmp.pddl")
                    simulator_learned = SequentialSimulator(problem=domain_learned)
                    os.remove('tmp.pddl')
                    app_actions_learned = get_applicable_actions(simulator_learned,
                                                                 simulator_learned.get_initial_state(),
                                                                 domain_learned,
                                                                 ground_actions)

                    # Update applicability statistics
                    for op in operators:
                        operators[op]['applicability']['tp'] += len(app_actions_learned[op].intersection(app_actions_ref[op]))
                        operators[op]['applicability']['fp'] += len(app_actions_learned[op] - (app_actions_ref[op]))
                        operators[op]['applicability']['fn'] += len(app_actions_ref[op] - (app_actions_learned[op]))

                    # Update consequentiality statistics
                    if applicable_learn and applicable_ref:  # if the action is applicable in both the learned and reference models
                        operators[a.action.name]['consequentiality']['tp'] += len((lnext_learned - lprev)
                                                                    .intersection((lnext_ref - lprev)))
                        operators[a.action.name]['consequentiality']['fp'] += len((lnext_learned - lprev) - lnext_ref)
                        operators[a.action.name]['consequentiality']['fn'] += len((lnext_learned.intersection(lprev)) - lnext_ref)

                    lprev = lnext_ref

            bar()  # update progress bar

    # Compute operator applicability precision and recall
    all_tp_app = 0
    all_fp_app = 0
    all_fn_app = 0
    all_precision_app = []
    all_recall_app = []
    for op in operators:
        if (operators[op]['applicability']['tp'] + operators[op]['applicability']['fp']) == 0:
            if operators[op]['applicability']['fn'] == 0:
                warnings.warn(f"Operator {op} has never been observed, setting predicted "
                              f"applicability precision to None.")
                operators[op]['applicability']['precision'] = None
            else:
                warnings.warn(f"Operator {op} has been observed but never predicted positive, "
                              f"setting predicted applicability precision to 1.")
                operators[op]['applicability']['precision'] = 1.
        else:
            operators[op]['applicability']['precision'] = ((operators[op]['applicability']['tp'])
                                                           / (operators[op]['applicability']['tp']
                                                              + operators[op]['applicability']['fp']))

        if (operators[op]['applicability']['tp'] + operators[op]['applicability']['fn']) == 0:
            if operators[op]['applicability']['fp'] == 0:
                warnings.warn(f"Operator {op} has never been observed, setting predicted "
                              f"applicability recall to None.")
                operators[op]['applicability']['recall'] = None
            else:
                warnings.warn(f"Operator {op} has been observed but never predicted negative, "
                              f"setting predicted applicability recall to 1.")
                operators[op]['applicability']['recall'] = 1.
        else:
            operators[op]['applicability']['recall'] = ((operators[op]['applicability']['tp'])
                                                        / (operators[op]['applicability']['tp']
                                                           + operators[op]['applicability']['fn']))
        all_tp_app += operators[op]['applicability']['tp']
        all_fp_app += operators[op]['applicability']['fp']
        all_fn_app += operators[op]['applicability']['fn']
        all_precision_app.append(operators[op]['applicability']['precision'])
        all_recall_app.append(operators[op]['applicability']['recall'])

    # Compute overall applicability precision and recall
    overall_precision_app = all_tp_app / (all_tp_app + all_fp_app)
    overall_recall_app = all_tp_app / (all_tp_app + all_fn_app)

    # Compute operator consequentiality precision and recall
    all_tp_cons = 0
    all_fp_cons = 0
    all_fn_cons = 0
    all_precision_cons = []
    all_recall_cons = []
    for op in operators:
        if (operators[op]['consequentiality']['tp'] + operators[op]['consequentiality']['fp']) == 0:
            if operators[op]['consequentiality']['fn'] == 0:
                warnings.warn(f"Operator {op} has never been observed, setting predicted "
                              f"effects precision to None.")
                operators[op]['consequentiality']['precision'] = None
            else:
                operators[op]['consequentiality']['precision'] = 1.
        else:
            operators[op]['consequentiality']['precision'] = ((operators[op]['consequentiality']['tp'])
                                                              / (operators[op]['consequentiality']['tp']
                                                                 + operators[op]['consequentiality']['fp']))

        if (operators[op]['consequentiality']['tp'] + operators[op]['consequentiality']['fn']) == 0:
            if operators[op]['consequentiality']['fp'] == 0:
                warnings.warn(f"Operator {op} has never been observed, setting predicted "
                              f"effects recall to None.")
                operators[op]['consequentiality']['recall'] = None
            else:
                operators[op]['consequentiality']['recall'] = 1.
        else:
            operators[op]['consequentiality']['recall'] = ((operators[op]['consequentiality']['tp'])
                                                           / (operators[op]['consequentiality']['tp']
                                                              + operators[op]['consequentiality']['fn']))
        all_tp_cons += operators[op]['consequentiality']['tp']
        all_fp_cons += operators[op]['consequentiality']['fp']
        all_fn_cons += operators[op]['consequentiality']['fn']
        all_precision_cons.append(operators[op]['consequentiality']['precision'])
        all_recall_cons.append(operators[op]['consequentiality']['recall'])

    # Remove undefined metrics (e.g. predicted effects (aka consequentiality) for operators that have
    # never been observed, or were never applicable in the learned model).
    all_precision_cons = [p for p in all_precision_cons if p is not None]
    all_recall_cons = [p for p in all_recall_cons if p is not None]

    # Compute overall consequentiality precision and recall
    overall_precision_cons = all_tp_cons / (all_tp_cons + all_fp_cons)
    overall_recall_cons = all_tp_cons / (all_tp_cons + all_fn_cons)

    return {
        'applicability': {
            'avg_precision': np.mean(all_precision_app),
            'avg_recall': np.mean(all_recall_app),
            'overall_precision': overall_precision_app,
            'overall_recall': overall_recall_app,
        },
        'predicted_effects': {
            'avg_precision': np.mean(all_precision_cons),
            'avg_recall': np.mean(all_recall_cons),
            'overall_precision': overall_precision_cons,
            'overall_recall': overall_recall_cons,
        },
        **operators
    }


def validate_plans(model_path: str,
                   problem_plan_paths: Dict[str, str]) -> float:
    """
    Validate the plans for the given problems according to the given model.
    :param model_path: path to an action model
    :param problem_plan_paths: dictionary where keys are problem paths and values the respective plan paths.
    :return:
    """
    reader = PDDLReader()
    valid_plans = 0
    for problem_path, plan_path in problem_plan_paths.items():
        # Parse problem
        problem = reader.parse_problem(model_path, problem_path)
        # Parse plan
        plan = reader.parse_plan(problem, plan_path)

        with PlanValidator() as validator:
            result = validator.validate(problem, plan)
            if result.status == ValidationResultStatus.VALID:
                valid_plans += 1

    return valid_plans / len(problem_plan_paths)


def validate_plan(model_path: str,
                  problem_path: str,
                  plan_path: str,) -> float:
    reader = PDDLReader()
    # Parse problem
    problem = reader.parse_problem(model_path, problem_path)
    # Parse plan
    plan = reader.parse_plan(problem, plan_path)

    with PlanValidator() as validator:
        result = validator.validate(problem, plan)

    return result.status == ValidationResultStatus.VALID


def problem_solving(model_learn_path: str,
                    model_ref_path: str,
                    problem_paths: List[str],
                    timeout=60) -> Dict[str, Dict[str, float]]:
    """
    Solve the given problems by means of the given model and return the solving and false plan ratio
    according to a reference model.
    :param model_learn_path: learned model path
    :param model_ref_path: reference model path
    :param problem_paths: list of problem paths
    :param timeout: planner timeout
    :return: solving_ratio and false_plans ratio
    """
    reader = PDDLReader()
    DOWNWARD_SEARCH_CFG = 'let(hff,ff(),let(hcea,cea(),lazy_greedy([hff,hcea],preferred=[hff,hcea])))'
    HEUR_PLANNER_CFG = {
        'name': 'fast-downward',
        'params': dict(
            fast_downward_search_config=DOWNWARD_SEARCH_CFG,
            fast_downward_search_time_limit=f"{timeout}s"
        )}

    solving = 0
    false_plans = 0
    unsolvable = 0
    timed_out = 0
    # Solve the problem with the learned model
    for problem_path in problem_paths:

        logging.info(f"Solving problem {problem_path}")

        problem = reader.parse_problem(model_learn_path, problem_path)

        with contextlib.redirect_stdout(open(os.devnull, 'w')):
            with OneshotPlanner(
                    problem_kind=problem.kind,
                    **HEUR_PLANNER_CFG
            ) as planner:
                result = planner.solve(problem, timeout=timeout)
                plan = result.plan

        if plan is not None:  # neither solving_plan nor false_plan
            # Parse problem
            problem_ref = reader.parse_problem(model_ref_path, problem_path)

            PDDLWriter(problem_ref).write_plan(plan, 'tmp')

            if validate_plan(model_ref_path, problem_path, 'tmp'):
                logging.debug(f"Solution plan for problem {problem_path} is valid")
                solving += 1
            else:
                logging.debug(f"Solution plan for problem {problem_path} is not valid")
                false_plans += 1

            os.remove('tmp')
        else:
            if result.status == PlanGenerationResultStatus.TIMEOUT:
                logging.debug(f"No solution plan found for problem {problem_path} within {timeout} seconds")
                timed_out += 1
            elif result.status in [PlanGenerationResultStatus.UNSOLVABLE_PROVEN, PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY]:
                logging.debug(f"Problem is not solvable")
                unsolvable += 1

    return {"problem_solving":
        {
            'solving_ratio': solving / len(problem_paths),
            'false_plans_ratio': false_plans / len(problem_paths),
            'unsolvable_ratio': unsolvable / len(problem_paths),
            'timed_out': timed_out / len(problem_paths)
        }
    }


def solving_eval(model_path: str,
                 problem_plan_paths: Dict[str, str]) -> Dict[str, float]:
    return {'validity': validate_plans(model_path, problem_plan_paths)}


if __name__ == '__main__':

    approaches = ['SAM', 'OffLAM', 'ROSAME', 'NOLAM']

    SYNTACTIC_EVAL = True
    PREDICTIVE_EVAL = True
    SOLVING_EVAL = True
    PROBLEM_SOLVING_DIR = 'problems/solving'

    logging.basicConfig(
        # filename='out.log',
        level=logging.INFO
    )

    metrics_syn = metrics_pred = metrics_solv = dict()

    MAX_TRACES_APPLICABILITY = 100  # lower this for faster debugging
    MAX_PLANNING_SEC = 60  # timeout for problem-solving metrics

    for approach in approaches:
        for run_dir in [d for d in os.listdir(f'../res/{approach}') if '.' not in d]:

            # Read existing metrics (e.g. CPU time)
            with open(f"../res/{approach}/{run_dir}/metrics.json", 'r') as f:
                domain_metrics = json.load(f)

            for model_learned_file in [d for d in os.listdir(f'../res/{approach}/{run_dir}')
                                       if '.pddl' in d]:

                domain = model_learned_file.split('_')[0]

                if domain not in []:
                    model_ref_path = f"../benchmarks/domains/{domain}.pddl"

                    traj_files = sorted(os.listdir(f"../benchmarks/trajectories/applicability/{domain}"),
                                        key=lambda x: int(x.split('/')[-1].split('_')[0]))

                    prob_files = sorted(os.listdir(f"../benchmarks/{PROBLEM_SOLVING_DIR}/{domain}"),
                                        key=lambda x: int(x.split('/')[-1].split('_')[0]))
                    trajs_probs_path = {
                        f"../benchmarks/trajectories/applicability/{domain}/{t}":
                            f"../benchmarks/problems/applicability/{domain}/{t.replace('_traj', '_prob.pddl')}"
                        for k, t in enumerate(traj_files) if k < MAX_TRACES_APPLICABILITY
                    }
                    prob_paths = [f"../benchmarks/{PROBLEM_SOLVING_DIR}/{domain}/{p}" for p in prob_files]

                    model_learned_path = f'../res/{approach}/{run_dir}/{model_learned_file}'

                    if SYNTACTIC_EVAL:
                        metrics_syn = syntactic_eval(model_learned_path, model_ref_path)

                    if PREDICTIVE_EVAL:
                        metrics_pred = predictive_eval(model_learned_path, model_ref_path, trajs_probs_path)

                    if SOLVING_EVAL:
                        metrics_solv = problem_solving(model_learned_path,
                                                       model_ref_path,
                                                       prob_paths,
                                                       timeout=MAX_PLANNING_SEC)

                    domain_metrics[domain] = {
                        **domain_metrics[domain],
                        **metrics_syn,
                        **metrics_pred,
                        **metrics_solv
                    }

            # Extend existing metrics
            with open(f"../res/{approach}/{run_dir}/metrics.json", 'w') as f:
                json.dump(domain_metrics, f, indent=3)

            # Save a reduced set of metrics for later plot
            df_metrics = []
            for d in domain_metrics:
                df_metrics.append({
                    'domain': d,
                    'syn precision': domain_metrics[d]['syntactic']['overall_precision']
                    if 'syntactic' in domain_metrics[d] else None,
                    'syn recall': domain_metrics[d]['syntactic']['overall_recall']
                    if 'syntactic' in domain_metrics[d] else None,
                    'app precision': domain_metrics[d]['applicability']['avg_precision']
                    if 'applicability' in domain_metrics[d] else None,
                    'app recall': domain_metrics[d]['applicability']['avg_recall']
                    if 'applicability' in domain_metrics[d] else None,
                    'predicted_effects precision': domain_metrics[d]['predicted_effects']['avg_precision']
                    if 'predicted_effects' in domain_metrics[d] else None,
                    'predicted_effects recall': domain_metrics[d]['predicted_effects']['avg_recall']
                    if 'predicted_effects' in domain_metrics[d] else None,
                    'solving_ratio': domain_metrics[d]['problem_solving']['solving_ratio']
                    if 'problem_solving' in domain_metrics[d] else None,
                    'false_plans_ratio': domain_metrics[d]['problem_solving']['false_plans_ratio']
                    if 'problem_solving' in domain_metrics[d] else None,
                })
            pd.DataFrame(df_metrics).to_excel(f"../res/{approach}/{run_dir}/metrics.xlsx",
                                              index=False, float_format='%.2f')



from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import Domain, Observation
from pytest import fixture

from sam_learning.core import EnvironmentSnapshot
from tests.consts import (
    BARMAN_ERROR_IN_PRODUCTION_PROBLEM_PATH,
    BARMAN_ERROR_IN_PRODUCTION_TRAJECTORY_PATH,
    BARMAN_ERROR_IN_PRODUCTION_DOMAIN_PATH,
)


@fixture()
def barman_domain() -> Domain:
    return DomainParser(BARMAN_ERROR_IN_PRODUCTION_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def elevators_environment_snapshot(elevators_domain: Domain):
    return EnvironmentSnapshot(partial_domain=elevators_domain)


@fixture()
def minecraft_environment_snapshot(minecraft_domain: Domain):
    return EnvironmentSnapshot(partial_domain=minecraft_domain)


@fixture()
def barman_environment_snapshot(barman_domain: Domain):
    return EnvironmentSnapshot(partial_domain=barman_domain)


@fixture()
def barman_observation(barman_domain: Domain) -> Observation:
    barman_problem = ProblemParser(
        problem_path=BARMAN_ERROR_IN_PRODUCTION_PROBLEM_PATH, domain=barman_domain
    ).parse_problem()
    return TrajectoryParser(barman_domain, barman_problem).parse_trajectory(BARMAN_ERROR_IN_PRODUCTION_TRAJECTORY_PATH)


def test_create_state_discrete_snapshot_creates_a_snapshot_with_negative_predicates_set_that_is_not_empty(
    elevators_environment_snapshot: EnvironmentSnapshot, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    initial_state = observation_component.previous_state
    observed_objects = elevators_observation.grounded_objects
    predicates = elevators_environment_snapshot._create_state_discrete_snapshot(
        state=initial_state, relevant_objects=observed_objects
    )
    negative_predicates = [p for p in predicates if not p.is_positive]
    assert len(negative_predicates) > 0


def test_create_state_discrete_snapshot_creates_a_snapshot_with_correct_object_types_for_objects_with_type_hierarchy_in_barman_domain(
    barman_environment_snapshot: EnvironmentSnapshot, barman_observation: Observation
):
    observation_component = barman_observation.components[0]
    pre_state = observation_component.next_state
    observed_objects = barman_observation.grounded_objects
    predicates = barman_environment_snapshot._create_state_discrete_snapshot(
        state=pre_state, relevant_objects=observed_objects
    )
    assert "(contains shot1 - shot ingredient1 - ingredient)" in [str(p) for p in predicates]
    assert "(contains shot1 - container ingredient1 - beverage)" not in [str(p) for p in predicates]


def test_create_state_discrete_snapshot_creates_a_snapshot_with_negative_and_positive_predicates(
    elevators_environment_snapshot: EnvironmentSnapshot, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    initial_state = observation_component.previous_state
    observed_objects = elevators_observation.grounded_objects
    predicates = elevators_environment_snapshot._create_state_discrete_snapshot(
        state=initial_state, relevant_objects=observed_objects
    )
    positive_predicates_str = set([p.untyped_representation for p in predicates if p.is_positive])
    negative_predicates_str = set([p.untyped_representation for p in predicates if not p.is_positive])
    assert positive_predicates_str.intersection(negative_predicates_str) == set()


def test_create_state_discrete_snapshot_creates_a_snapshot_with_positive_predicates_covering_all_state_predicates(
    elevators_environment_snapshot: EnvironmentSnapshot, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    initial_state = observation_component.previous_state
    observed_objects = elevators_observation.grounded_objects
    predicates = elevators_environment_snapshot._create_state_discrete_snapshot(
        state=initial_state, relevant_objects=observed_objects
    )
    positive_predicates = [p for p in predicates if p.is_positive]
    state_predicates = set()
    for predicates in initial_state.state_predicates.values():
        state_predicates.update(predicates)

    positive_predicates_str = set([p.untyped_representation for p in positive_predicates])
    state_predicates_str = set([p.untyped_representation for p in state_predicates])
    assert len(positive_predicates_str.intersection(state_predicates_str)) == len(state_predicates_str)
    assert len(positive_predicates_str.intersection(state_predicates_str)) == len(positive_predicates_str)


def test_create_snapshot_when_should_not_include_all_objects_creates_lower_number_of_predicates_than_when_with_all_objects(
    elevators_environment_snapshot: EnvironmentSnapshot, elevators_observation: Observation, elevators_domain: Domain
):
    observation_component = elevators_observation.components[0]
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    current_action = observation_component.grounded_action_call
    observed_objects = elevators_observation.grounded_objects
    elevators_environment_snapshot.create_triplet_snapshot(
        previous_state=previous_state,
        next_state=next_state,
        current_action=current_action,
        observation_objects=observed_objects,
    )
    num_predicates_without_all_objects = len(elevators_environment_snapshot.previous_state_predicates)

    all_types = [pddl_type for pddl_type in elevators_domain.types if pddl_type != "object"]
    elevators_environment_snapshot.create_triplet_snapshot(
        previous_state=previous_state,
        next_state=next_state,
        current_action=current_action,
        observation_objects=observed_objects,
        specific_types=all_types,
    )
    num_predicates_with_all_objects = len(elevators_environment_snapshot.previous_state_predicates)

    assert num_predicates_without_all_objects < num_predicates_with_all_objects


def test_create_numeric_state_snapshot_does_not_exclude_numeric_functions_that_contain_constant_parameters(
    minecraft_observation: Observation, minecraft_domain: Domain, minecraft_environment_snapshot: EnvironmentSnapshot
):
    tested_component = minecraft_observation.components[13]
    state_to_test = tested_component.previous_state
    test_action = tested_component.grounded_action_call
    observed_objects = {**minecraft_observation.grounded_objects, **minecraft_domain.constants}
    snapshot = minecraft_environment_snapshot.create_numeric_state_snapshot(
        state=state_to_test, current_action=test_action, observation_objects=observed_objects
    )
    assert len(snapshot.keys()) == 8


def test_create_triplet_snapshot_does_not_exclude_numeric_functions_that_contain_constant_parameters(
    minecraft_observation: Observation, minecraft_domain: Domain, minecraft_environment_snapshot: EnvironmentSnapshot
):
    tested_component = minecraft_observation.components[13]
    previous_state = tested_component.previous_state
    test_action = tested_component.grounded_action_call
    next_state = tested_component.next_state
    observed_objects = minecraft_observation.grounded_objects
    minecraft_environment_snapshot.create_triplet_snapshot(
        previous_state=previous_state,
        next_state=next_state,
        current_action=test_action,
        observation_objects=observed_objects,
    )
    assert len(minecraft_environment_snapshot.previous_state_functions.keys()) == 8



from enum import Enum


class LearningAlgorithmType(Enum):
    sam_learning = 1
    esam_learning = 2
    numeric_sam = 3
    plan_miner = 5
    ma_sam = 7
    conditional_sam = 9
    universal_sam = 10
    online_nsam = 13
    incremental_nsam = 19

    # learning algorithms relating to fault repair not to be used for POL
    oblique_tree = 11
    extended_svc = 12

    # To remove
    naive_nsam = 15
    naive_polysam = 16
    ma_sam_plus = 25


class SolverType(Enum):
    fast_downward = 1
    metric_ff = 2
    enhsp = 3
    fast_forward = 4


class SolutionOutputTypes(Enum):
    ok = 1
    no_solution = 2
    timeout = 3
    not_applicable = 4
    goal_not_achieved = 5
    solver_error = 6


class NegativePreconditionPolicy(Enum):
    no_remove = (1,)  # keep all learned preconditions
    soft = (2,)
    hard = (3,)  # no negative preconditions are allowed
    # allow only negative preconditions deducted form cnf solution


NUMERIC_PRECISION = "NUMERIC_PRECISION"



from distutils.core import setup


with open("README.md") as f:
    long_description = f.read()

setup(name='n-sam',
      version='1.2.0',
      python_requires=">=3.8",
      description='Numeric Safe Action Model Learning algorithm implementation in Python',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Argaman Mordoch',
      packages=["sam_learning", "solvers", "observations"]
     )


"""Module to match action call parameters to the functions observed in the state."""
import itertools
import logging
from itertools import permutations
from typing import List, Dict, Union

from pddl_plus_parser.models import Domain, PDDLFunction, ActionCall, Action


class NumericFunctionMatcher:
    """Class that matched numeric functions to the actions according to their parameters."""

    matcher_domain: Domain
    logger: logging.Logger

    def __init__(self, domain: Domain):
        self.logger = logging.getLogger(__name__)
        self.matcher_domain = domain

    @staticmethod
    def create_possible_function_signatures(action_parameters: List[str], numeric_function: PDDLFunction) -> List[str]:
        """Create all possible grounded calls of the function.

        Note: Assuming the input function has less or the same number of parameters as the action.

        :param action_parameters: the parameters that the action was called with.
        :param numeric_function:the numeric function that is defined in the domain.
        :return: possible unsigned representations that will be used to map the correct state function.
        """
        num_function_params = len(numeric_function.signature)
        possible_parameters = permutations(action_parameters, num_function_params)
        possible_untyped_signatures = []
        for permutation in possible_parameters:
            untyped_signature_str = " ".join(permutation)
            possible_untyped_signatures.append(f"({numeric_function.name} {untyped_signature_str})")

        return possible_untyped_signatures

    def lift_matched_parameters(
        self, executed_action: Action, grounded_call_parameters: List[str], grounded_function: PDDLFunction
    ) -> List[PDDLFunction]:
        """Matches the parameters of the function to the lifted action definition.

        :param executed_action: the lifted action that was called in the observation.
        :param grounded_call_parameters: the observed action call objects.
        :param grounded_function: the observed numeric function.
        :return: the lifted function definition with the value set according to the observation.
        """
        self.logger.debug("Starting to lift the matched parameters!")
        object_to_parameter_mapping = [
            {grounded_obj: lifted_param}
            for grounded_obj, lifted_param in zip(grounded_call_parameters, executed_action.parameter_names)
        ]
        object_to_parameter_mapping.extend([{const: const} for const in self.matcher_domain.constants])
        lifted_params_and_consts = {
            **executed_action.signature,
            **{const_name: const.type for const_name, const in self.matcher_domain.constants.items()},
        }
        grounded_objects = list(grounded_function.signature.keys())
        map_permutations = list(itertools.product(object_to_parameter_mapping, repeat=len(grounded_objects)))
        lifted_matches = []
        for permutation in map_permutations:
            possible_grounded_execution = [list(item.keys())[0] for item in permutation]
            if possible_grounded_execution == grounded_objects:
                lifted_matched_signature = {}
                for item in permutation:
                    lifted_param_name = item[list(item.keys())[0]]
                    lifted_matched_signature[lifted_param_name] = lifted_params_and_consts[lifted_param_name]

                lifted_state_function = PDDLFunction(name=grounded_function.name, signature=lifted_matched_signature)
                lifted_state_function.set_value(grounded_function.value)
                lifted_matches.append(lifted_state_function)

        return lifted_matches

    def match_state_functions(
        self, action_call: ActionCall, grounded_state_fluents: Dict[str, PDDLFunction]
    ) -> Dict[str, PDDLFunction]:
        """Match the state functions to the action according to the grounded object names.

        :param action_call: the action that was called in the observation.
        :param grounded_state_fluents: the numeric state fluents.
        :return: a dictionary containing the lifted numeric state function with their assigned values.
        """
        self.logger.debug(f"Starting to search for matches for the grounded action - {str(action_call)}")
        possible_matches = {}
        for domain_function in self.matcher_domain.functions.values():
            if len(domain_function.signature) > len(action_call.parameters) and len(self.matcher_domain.constants) == 0:
                self.logger.debug(f"Function - {domain_function.name} has too many parameters, skipping.")
                continue

            if len(domain_function.signature) == 0:
                grounded_function = grounded_state_fluents[domain_function.untyped_representation]
                self.logger.debug(
                    f"Function {domain_function.name} has zero parameters. " f"Instantly considering as match."
                )
                matched_function = PDDLFunction(name=domain_function.name, signature={})
                matched_function.set_value(grounded_function.value)
                possible_matches[domain_function.untyped_representation] = matched_function
                continue

            call_parameters = action_call.parameters + list(self.matcher_domain.constants.keys())
            possible_function_signatures = self.create_possible_function_signatures(call_parameters, domain_function)
            for unsigned_representation in possible_function_signatures:
                if unsigned_representation in grounded_state_fluents:
                    self.logger.debug(
                        f"found a possible match to the action " f"{action_call.name} - {unsigned_representation}"
                    )

                    matched_lifted_functions = self.lift_matched_parameters(
                        executed_action=self.matcher_domain.actions[action_call.name],
                        grounded_call_parameters=action_call.parameters,
                        grounded_function=grounded_state_fluents[unsigned_representation],
                    )
                    for match in matched_lifted_functions:
                        possible_matches[match.untyped_representation] = match

        return possible_matches



from sam_learning.learners import NumericSAMLearner, IncrementalNumericSAMLearner
from sam_learning.learners.baseline_learners.naive_numeric_sam import NaivePolynomialSAMLearning
from utilities import LearningAlgorithmType

DEFAULT_SPLIT = 5

NUMERIC_ALGORITHMS = [
    LearningAlgorithmType.numeric_sam,
    LearningAlgorithmType.plan_miner,
    LearningAlgorithmType.naive_nsam,
    LearningAlgorithmType.naive_polysam,
    LearningAlgorithmType.incremental_nsam,
]

DEFAULT_NUMERIC_TOLERANCE = 0.1

MAX_SIZE_MB = 5


NUMERIC_SAM_ALGORITHM_VERSIONS = {
    LearningAlgorithmType.numeric_sam: NumericSAMLearner,
    LearningAlgorithmType.naive_nsam: NaivePolynomialSAMLearning,
    LearningAlgorithmType.incremental_nsam: IncrementalNumericSAMLearner,
}



"""Module to validate the correctness of the learned action models that were generated."""
import csv
import logging
import re
import time
from pathlib import Path
from typing import Dict, List, Any, Optional, Union

from pddl_plus_parser.models import Observation, MultiAgentObservation

from solvers import FastDownwardSolver, MetricFFSolver, ENHSPSolver, FFADLSolver
from utilities import LearningAlgorithmType, SolverType, SolutionOutputTypes, NegativePreconditionPolicy, MappingElement, MacroActionParser
from validators.common import AGGREGATED_SOLVING_FIELDS
from validators.validator_script_data import VALID_PLAN, INAPPLICABLE_PLAN, GOAL_NOT_REACHED, run_validate_script

SOLVER_TYPES = {
    SolverType.fast_downward: FastDownwardSolver,
    SolverType.metric_ff: MetricFFSolver,
    SolverType.enhsp: ENHSPSolver,
    SolverType.fast_forward: FFADLSolver,
}

SOLVING_STATISTICS = [
    "learning_algorithm",
    "num_trajectories",
    "num_trajectory_triplets",
    "policy",
    "learning_time",
    "solving_time",
    "solver",
    "ok",
    "no_solution",
    "timeout",
    "solver_error",
    "not_applicable",
    "goal_not_achieved",
    "validated_against_expert_plan",
    "not_validated_against_expert_plan",
    "problems_ok",
    "problems_no_solution",
    "problems_timeout",
    "problems_solver_error",
    "problems_not_applicable",
    "problems_goal_not_achieved",
    "problems_validated_against_expert_plan",
    "problems_not_validated_against_expert_plan",
    "percent_ok",
    "percent_no_solution",
    "percent_timeout",
    "percent_solver_error",
    "percent_not_applicable",
    "percent_goal_not_achieved",
    "percent_validated_against_expert_plan",
    "percent_not_validated_against_expert_plan",
]

DEBUG_STATISTICS = [
    "problems_ok",
    "problems_no_solution",
    "problems_timeout",
    "problems_solver_error",
    "problems_not_applicable",
    "problems_goal_not_achieved",
    "problems_validated_against_expert_plan",
    "problems_not_validated_against_expert_plan",
]

VALIDATED_AGAINST_EXPERT_PLAN = "validated_against_expert_plan"
NOT_VALIDATED_AGAINST_EXPERT_PLAN = "not_validated_against_expert_plan"

NUMERIC_STATISTICS_LABELS = [
    *[solution_type.name for solution_type in SolutionOutputTypes],
    VALIDATED_AGAINST_EXPERT_PLAN,
    NOT_VALIDATED_AGAINST_EXPERT_PLAN,
]


ACTION_LINE_REGEX = r"(\[\d+, \d+\]): \(.*\)"


class DomainValidator:
    """Validates that the learned domain can create plans.

    Note:
        There is no validation on the plan content since the learning process is safe.
    """

    logger: logging.Logger
    solving_stats: List[Dict[str, Any]]
    aggregated_solving_stats: List[Dict[str, Any]]
    learning_algorithm: LearningAlgorithmType
    working_directory_path: Path
    reference_domain_path: Path
    results_dir_path: Path
    problem_prefix: str

    def __init__(
        self, working_directory_path: Path, learning_algorithm: LearningAlgorithmType, reference_domain_path: Path, problem_prefix: str = "pfile",
    ):
        self.logger = logging.getLogger(__name__)
        self.solving_stats = []
        self.aggregated_solving_stats = []
        self.learning_algorithm = learning_algorithm
        self.results_dir_path = working_directory_path / "results_directory"
        self.reference_domain_path = reference_domain_path
        self.problem_prefix = problem_prefix
        self.working_directory_path = working_directory_path

    @staticmethod
    def _clear_plans(test_set_directory: Path) -> None:
        """Clears the plan filed from the directory.

        :param test_set_directory: the path to the directory containing the plans.
        """
        for solver_output_path in test_set_directory.glob("*.solution"):
            solver_output_path.unlink(missing_ok=True)

    def _validate_against_expert_plan(
        self, solution_file_path: Path, problem_file_path: Path, iteration_statistics: Dict[str, Union[int, List[str]]], tested_domain_path: Path
    ) -> None:
        """Validates that the expert solution can be used against the learned domain.

        :param solution_file_path: the path to the solution file.
        :param problem_file_path: the path to the problem file.
        :param iteration_statistics: the statistics of the current iteration.
        :param tested_domain_path: the path to the domain file to validate against.
        """
        validation_file_path = run_validate_script(
            domain_file_path=tested_domain_path, problem_file_path=problem_file_path, solution_file_path=solution_file_path
        )
        with open(validation_file_path, "r") as validation_file:
            validation_file_content = validation_file.read()
            if VALID_PLAN in validation_file_content:
                self.logger.info("The expert plan is valid on the learned domain.")
                iteration_statistics[VALIDATED_AGAINST_EXPERT_PLAN] += 1
                iteration_statistics[f"problems_{VALIDATED_AGAINST_EXPERT_PLAN}"].append(problem_file_path.name)

            else:
                self.logger.info("The expert plan is not valid on the learned domain.")
                iteration_statistics[NOT_VALIDATED_AGAINST_EXPERT_PLAN] += 1
                iteration_statistics[f"problems_{NOT_VALIDATED_AGAINST_EXPERT_PLAN}"].append(problem_file_path.name)

        validation_file_path.unlink(missing_ok=True)

    def _validate_solution_content(
        self, solution_file_path: Path, problem_file_path: Path, iteration_statistics: Dict[str, Union[int, List[str]]]
    ) -> None:
        """Validates that the solution file contains a valid plan.

        :param solution_file_path: the path to the solution file.
        :param problem_file_path: the path to the problem file.
        :param iteration_statistics: the statistics of the current iteration.
        """
        validation_file_path = run_validate_script(
            domain_file_path=self.reference_domain_path, problem_file_path=problem_file_path, solution_file_path=solution_file_path
        )

        with open(validation_file_path, "r", encoding="utf-8") as validation_file:
            validation_file_content = validation_file.read()
            if VALID_PLAN in validation_file_content:
                self.logger.info("The plan is valid.")
                iteration_statistics["ok"] += 1
                iteration_statistics["problems_ok"].append(problem_file_path.name)

            elif INAPPLICABLE_PLAN in validation_file_content:
                self.logger.info("The plan is not applicable.")
                iteration_statistics["not_applicable"] += 1
                iteration_statistics["problems_not_applicable"].append(problem_file_path.name)

            elif GOAL_NOT_REACHED in validation_file_content:
                self.logger.info("The plan did not reach the required goal.")
                iteration_statistics["goal_not_achieved"] += 1
                iteration_statistics["problems_goal_not_achieved"].append(problem_file_path.name)

        validation_file_path.unlink(missing_ok=True)

    @staticmethod
    def _extract_num_triplets(used_observations: Union[List[Observation], List[MultiAgentObservation], List[Path]] = None) -> int:
        """Extracts the number of trajectory triplets from the observations.

        :param used_observations: the observations used to generate the plans.
        :return: the number of trajectory triplets in the used observations.
        """
        if not used_observations:
            return 0

        for observation in used_observations:
            if isinstance(observation, Path):
                with open(observation, "r") as observation_file:
                    num_operators = 0
                    for line in observation_file.readlines():
                        match = re.match(ACTION_LINE_REGEX, line)
                        num_operators = num_operators + 1 if match else num_operators

                    return num_operators

        num_triplets = sum([len(observation.components) for observation in used_observations])
        return num_triplets

    @staticmethod
    def _calculate_expert_validation_statistics(solving_stats: Dict[str, Any]) -> None:
        """Calculates the expert validation statistics.

        :param solving_stats: the solving statistics.
        """
        total_validated = sum([solving_stats[statistic] for statistic in [VALIDATED_AGAINST_EXPERT_PLAN, NOT_VALIDATED_AGAINST_EXPERT_PLAN]])
        total_validated = total_validated if total_validated > 0 else 1

        for statistic in [VALIDATED_AGAINST_EXPERT_PLAN, NOT_VALIDATED_AGAINST_EXPERT_PLAN]:
            solving_stats[f"percent_{statistic}"] = 100 * (solving_stats[statistic] / total_validated)

    def _calculate_solving_percentages(self, solving_stats: Dict[str, Any]) -> None:
        """Calculates the percentage of solved problems.

        :param solving_stats:  the solving statistics.
        """
        total_problems = sum([solving_stats[statistic] for statistic in AGGREGATED_SOLVING_FIELDS])
        total_problems = total_problems if total_problems > 0 else 1
        for statistic in AGGREGATED_SOLVING_FIELDS:
            percentage_statistic = 100 * (solving_stats[statistic] / total_problems)
            solving_stats[f"percent_{statistic}"] = percentage_statistic
            self.logger.info(f"{statistic} percentage: {percentage_statistic:.2f}%")

    @staticmethod
    def adapt_solution_file(solution_path: Path, mapping: Dict[str, MappingElement]):
        """
        Post-processing solution files with macro actions and adapt them to something validators can work with.
        That essentially means replacing macro action lines with its consisting solo actions lines.

        :param solution_path: the path to the solution file output from the learned domain.
        :param mapping: the macro actions mapping of the learned domain.
        """
        with open(solution_path, "r") as file:
            lines = file.readlines()

        new_lines = []
        for line in lines:
            extracted_lines = MacroActionParser.extract_actions_from_macro_action(action_line=line, mapper=mapping)
            new_lines.extend([extracted_line if extracted_line.endswith("\n") else f"{extracted_line}\n" for extracted_line in extracted_lines])

        with open(solution_path, "w") as file:
            file.writelines(new_lines)

    def validate_domain(
        self,
        tested_domain_file_path: Path,
        test_set_directory_path: Optional[Path] = None,
        used_observations: Union[List[Union[Observation, MultiAgentObservation]], List[Path]] = None,
        tolerance: float = 0.01,
        timeout: int = 5,
        learning_time: float = 0,
        solvers_portfolio: List[SolverType] = None,
        preconditions_removal_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
        mapping: Dict[str, MappingElement] = None,
    ) -> None:
        """Validates that using the input domain problems can be solved.

        :param tested_domain_file_path: the path of the domain that was learned using POL.
        :param test_set_directory_path: the path to the directory containing the test set problems.
        :param used_observations: the observations that were used to learn the domain.
        :param timeout: the timeout for the solver.
        :param tolerance: the numeric tolerance to use.
        :param learning_time: the time it took to learn the domain (in seconds).
        :param solvers_portfolio: the solvers to use for the validation, can be one or more and each will try to solve each planning problem at most once.
        :param preconditions_removal_policy: the policy used to remove the negative discrete preconditions.
        :param mapping: the learned model mapper from macro action to mapping element.
        """
        num_triplets = self._extract_num_triplets(used_observations)
        solving_stats: Dict[str, Any] = {label: 0 for label in NUMERIC_STATISTICS_LABELS}
        for debug_statistic in DEBUG_STATISTICS:
            solving_stats[debug_statistic] = []

        self.logger.info("Solving the test set problems using the learned domain!")
        problem_solving_times = []
        for problem_path in test_set_directory_path.glob(f"{self.problem_prefix}*.pddl"):
            problem_solving_report = {}
            problem_solved = False
            problem_file_name = problem_path.stem
            for solver_type in solvers_portfolio:
                start_time = time.time()
                solver = SOLVER_TYPES[solver_type]()
                solver.solve_problem(
                    problems_directory_path=test_set_directory_path,
                    domain_file_path=tested_domain_file_path,
                    problem_file_path=problem_path,
                    tolerance=tolerance,
                    solving_timeout=timeout,
                    solving_stats=problem_solving_report,
                )
                end_time = time.time()
                if problem_solving_report[problem_file_name] == SolutionOutputTypes.ok.name:
                    problem_solving_times.append(end_time - start_time)  # time in seconds
                    problem_solved = True
                    solution_file_path = test_set_directory_path / f"{problem_file_name}.solution"

                    if mapping:
                        self.adapt_solution_file(solution_path=solution_file_path, mapping=mapping)

                    self._validate_solution_content(
                        solution_file_path=solution_file_path, problem_file_path=problem_path, iteration_statistics=solving_stats
                    )
                    break

            if problem_solved:
                continue

            expert_plan_path = self.working_directory_path / f"{problem_file_name}.solution"
            solving_outcome = problem_solving_report[problem_file_name]
            solving_stats[solving_outcome] += 1
            solving_stats[f"problems_{solving_outcome}"].append(problem_file_name)
            self._validate_against_expert_plan(
                solution_file_path=expert_plan_path,
                problem_file_path=problem_path,
                iteration_statistics=solving_stats,
                tested_domain_path=tested_domain_file_path,
            )

        solving_stats["solving_time"] = (
            sum(problem_solving_times) / len(problem_solving_times) if len(problem_solving_times) > 0 else 0
        )  # average time in seconds for all solved problems
        self._calculate_solving_percentages(solving_stats)
        self._calculate_expert_validation_statistics(solving_stats)
        num_trajectories = len(used_observations)
        self.solving_stats.append(
            {
                "learning_algorithm": self.learning_algorithm.name,
                "num_trajectories": num_trajectories,
                "policy": preconditions_removal_policy.name,
                "num_trajectory_triplets": num_triplets,
                "learning_time": learning_time,
                "solver": [solver.name for solver in solvers_portfolio],
                **solving_stats,
            }
        )
        self._clear_plans(test_set_directory_path)

    def write_statistics(self, fold_num: int, iteration: Optional[int] = None) -> None:
        """Writes the statistics of the learned model into a CSV file.

        :param fold_num: the index of the fold that is currently being tested.
        :param iteration: the index of the iteration that is currently being tested.
        """
        output_statistics_path = self.results_dir_path / (
            f"{self.learning_algorithm.name}"
            f"_problem_solving_stats_fold_{fold_num}"
            f"{f'_{iteration}_trajectories' if iteration is not None else ''}.csv"
        )
        with open(output_statistics_path, "wt", newline="") as csv_file:
            test_set_writer = csv.DictWriter(csv_file, fieldnames=SOLVING_STATISTICS)
            test_set_writer.writeheader()
            test_set_writer.writerows(self.solving_stats)

    def write_complete_joint_statistics(self, fold: Optional[int] = None) -> None:
        """Writes a statistics file containing all the folds combined data."""

        output_path = (
            self.results_dir_path / f"{self.learning_algorithm.name}_all_folds_solving_stats.csv"
            if fold is None
            else self.results_dir_path / f"{self.learning_algorithm.name}_problem_solving_stats_{fold}.csv"
        )
        with open(output_path, "wt", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=SOLVING_STATISTICS)
            writer.writeheader()
            writer.writerows(self.aggregated_solving_stats)

    def clear_statistics(self) -> None:
        """Clears the statistics so that each fold will have no relation to its predecessors."""
        self.aggregated_solving_stats.extend(self.solving_stats)
        self.solving_stats.clear()
